import asyncio
import base64
import functools
import json
import logging
import os
import re
import threading
import warnings
from unittest import mock

import pytest
import starlette
from starlette.authentication import (
    AuthCredentials,
    AuthenticationBackend,
    AuthenticationError,
    SimpleUser,
)
from starlette.exceptions import HTTPException
from starlette.middleware import Middleware
from starlette.middleware.authentication import AuthenticationMiddleware
from starlette.middleware.trustedhost import TrustedHostMiddleware
from starlette.testclient import TestClient

import sentry_sdk
from sentry_sdk import capture_message, get_baggage, get_traceparent
from sentry_sdk.consts import SPANDATA
from sentry_sdk.integrations.asgi import SentryAsgiMiddleware
from sentry_sdk.integrations.starlette import (
    StarletteIntegration,
)
from sentry_sdk.utils import parse_version
from tests.integrations.conftest import parametrize_test_configurable_status_codes

STARLETTE_VERSION = parse_version(starlette.__version__)

PICTURE = os.path.join(os.path.dirname(os.path.abspath(__file__)), "photo.jpg")

BODY_JSON = {"some": "json", "for": "testing", "nested": {"numbers": 123}}

BODY_FORM = """--fd721ef49ea403a6\r\nContent-Disposition: form-data; name="username"\r\n\r\nJane\r\n--fd721ef49ea403a6\r\nContent-Disposition: form-data; name="password"\r\n\r\nhello123\r\n--fd721ef49ea403a6\r\nContent-Disposition: form-data; name="photo"; filename="photo.jpg"\r\nContent-Type: image/jpg\r\nContent-Transfer-Encoding: base64\r\n\r\n{{image_data}}\r\n--fd721ef49ea403a6--\r\n""".replace(
    "{{image_data}}", str(base64.b64encode(open(PICTURE, "rb").read()))
)

PARSED_FORM = starlette.datastructures.FormData(
    [
        ("username", "Jane"),
        ("password", "hello123"),
        (
            "photo",
            starlette.datastructures.UploadFile(
                filename="photo.jpg",
                file=open(PICTURE, "rb"),
            ),
        ),
    ]
)


async def _mock_receive(msg):
    return msg


from starlette.templating import Jinja2Templates


def starlette_app_factory(middleware=None, debug=True):
    template_dir = os.path.join(
        os.getcwd(), "tests", "integrations", "starlette", "templates"
    )
    templates = Jinja2Templates(directory=template_dir)

    async def _homepage(request):
        1 / 0
        return starlette.responses.JSONResponse({"status": "ok"})

    async def _custom_error(request):
        raise Exception("Too Hot")

    async def _message(request):
        capture_message("hi")
        return starlette.responses.JSONResponse({"status": "ok"})

    async def _nomessage(request):
        return starlette.responses.JSONResponse({"status": "ok"})

    async def _message_with_id(request):
        capture_message("hi")
        return starlette.responses.JSONResponse({"status": "ok"})

    def _thread_ids_sync(request):
        return starlette.responses.JSONResponse(
            {
                "main": threading.main_thread().ident,
                "active": threading.current_thread().ident,
            }
        )

    async def _thread_ids_async(request):
        return starlette.responses.JSONResponse(
            {
                "main": threading.main_thread().ident,
                "active": threading.current_thread().ident,
            }
        )

    async def _render_template(request):
        capture_message(get_traceparent() + "\n" + get_baggage())

        template_context = {
            "request": request,
            "msg": "Hello Template World!",
        }
        if STARLETTE_VERSION >= (1,):
            return templates.TemplateResponse(
                request, "trace_meta.html", template_context
            )
        else:
            return templates.TemplateResponse("trace_meta.html", template_context)

    async def _body_json(request):
        await request.json()
        capture_message("hi")
        return starlette.responses.JSONResponse({"status": "ok"})

    async def _body_form(request):
        await request.form()
        capture_message("hi")
        return starlette.responses.JSONResponse({"status": "ok"})

    async def _body_raw(request):
        await request.body()
        capture_message("hi")
        return starlette.responses.JSONResponse({"status": "ok"})

    all_methods = [
        "CONNECT",
        "DELETE",
        "GET",
        "HEAD",
        "OPTIONS",
        "PATCH",
        "POST",
        "PUT",
        "TRACE",
    ]

    mounted_app = starlette.applications.Starlette(
        routes=[
            starlette.routing.Route("/nomessage", _nomessage, methods=all_methods),
        ],
    )

    app = starlette.applications.Starlette(
        debug=debug,
        routes=[
            starlette.routing.Route("/some_url", _homepage),
            starlette.routing.Route("/custom_error", _custom_error),
            starlette.routing.Route("/message", _message),
            starlette.routing.Route("/nomessage", _nomessage, methods=all_methods),
            starlette.routing.Route("/message/{message_id}", _message_with_id),
            starlette.routing.Route("/sync/thread_ids", _thread_ids_sync),
            starlette.routing.Route("/async/thread_ids", _thread_ids_async),
            starlette.routing.Route("/render_template", _render_template),
            starlette.routing.Route("/body/json", _body_json, methods=["POST"]),
            starlette.routing.Route("/body/form", _body_form, methods=["POST"]),
            starlette.routing.Route("/body/raw", _body_raw, methods=["POST"]),
            starlette.routing.Mount("/root", app=mounted_app),
        ],
        middleware=middleware,
    )

    return app


def async_return(result):
    f = asyncio.Future()
    f.set_result(result)
    return f


class BasicAuthBackend(AuthenticationBackend):
    async def authenticate(self, conn):
        if "Authorization" not in conn.headers:
            return

        auth = conn.headers["Authorization"]
        try:
            scheme, credentials = auth.split()
            if scheme.lower() != "basic":
                return
            decoded = base64.b64decode(credentials).decode("ascii")
        except (ValueError, UnicodeDecodeError):
            raise AuthenticationError("Invalid basic auth credentials")

        username, _, password = decoded.partition(":")

        # TODO: You'd want to verify the username and password here.

        return AuthCredentials(["authenticated"]), SimpleUser(username)


class AsyncIterator:
    def __init__(self, data):
        self.iter = iter(bytes(data, "utf-8"))

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            return bytes([next(self.iter)])
        except StopIteration:
            raise StopAsyncIteration


class SampleMiddleware:
    def __init__(self, app):
        self.app = app

    async def __call__(self, scope, receive, send):
        # only handle http requests
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        async def do_stuff(message):
            if message["type"] == "http.response.start":
                # do something here.
                pass

            await send(message)

        await self.app(scope, receive, do_stuff)


class SampleMiddlewareWithArgs(Middleware):
    def __init__(self, app, bla=None):
        self.app = app
        self.bla = bla


class SampleReceiveSendMiddleware:
    def __init__(self, app):
        self.app = app

    async def __call__(self, scope, receive, send):
        message = await receive()
        assert message
        assert message["type"] == "http.request"

        send_output = await send({"type": "something-unimportant"})
        assert send_output is None

        await self.app(scope, receive, send)


class SamplePartialReceiveSendMiddleware:
    def __init__(self, app):
        self.app = app

    async def __call__(self, scope, receive, send):
        message = await receive()
        assert message
        assert message["type"] == "http.request"

        send_output = await send({"type": "something-unimportant"})
        assert send_output is None

        async def my_receive(*args, **kwargs):
            pass

        async def my_send(*args, **kwargs):
            pass

        partial_receive = functools.partial(my_receive)
        partial_send = functools.partial(my_send)

        await self.app(scope, partial_receive, partial_send)


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_request_info_json_body(
    sentry_init, capture_events, capture_items, span_streaming
):
    sentry_init(
        traces_sample_rate=1.0,
        send_default_pii=True,
        integrations=[StarletteIntegration()],
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    starlette_app = starlette_app_factory()
    client = TestClient(starlette_app)

    if span_streaming:
        items = capture_items("event", "span")

        client.post(
            "/body/json",
            json=BODY_JSON,
            headers={
                "cookie": "yummy_cookie=choco; tasty_cookie=strawberry",
            },
        )

        (event,) = (item.payload for item in items if item.type == "event")
        assert event["request"]["cookies"] == {
            "tasty_cookie": "strawberry",
            "yummy_cookie": "choco",
        }
        assert event["request"]["data"] == BODY_JSON

        sentry_sdk.flush()
        spans = [item.payload for item in items if item.type == "span"]
        server_span = next(
            span for span in spans if span["attributes"]["sentry.op"] == "http.server"
        )

        assert json.loads(
            server_span["attributes"][SPANDATA.HTTP_REQUEST_BODY_DATA]
        ) == {"some": "json", "for": "testing", "nested": {"numbers": 123}}
    else:
        events = capture_events()

        client.post(
            "/body/json",
            json=BODY_JSON,
            headers={
                "cookie": "yummy_cookie=choco; tasty_cookie=strawberry",
            },
        )

        (event, transaction_event) = events

        assert event["request"]["cookies"] == {
            "tasty_cookie": "strawberry",
            "yummy_cookie": "choco",
        }
        assert event["request"]["data"] == BODY_JSON

        assert transaction_event["request"]["cookies"] == {
            "tasty_cookie": "strawberry",
            "yummy_cookie": "choco",
        }
        assert transaction_event["request"]["data"] == BODY_JSON


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_formdata_request_body(
    sentry_init, capture_events, capture_items, span_streaming
):
    sentry_init(
        traces_sample_rate=1.0,
        send_default_pii=True,
        max_request_body_size="always",
        integrations=[StarletteIntegration()],
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    starlette_app = starlette_app_factory()
    client = TestClient(starlette_app)

    if span_streaming:
        items = capture_items("event", "span")

        client.post(
            "/body/form",
            data=BODY_FORM.encode("utf-8"),
            headers={
                "content-type": "multipart/form-data; boundary=fd721ef49ea403a6",
            },
        )

        (event,) = (item.payload for item in items if item.type == "event")
        assert event["request"]["data"].keys() == PARSED_FORM.keys()
        assert event["request"]["data"]["username"] == PARSED_FORM["username"]
        assert event["request"]["data"]["password"] == "[Filtered]"
        assert event["request"]["data"]["photo"] == ""

        sentry_sdk.flush()
        spans = [item.payload for item in items if item.type == "span"]
        server_span = next(
            span for span in spans if span["attributes"]["sentry.op"] == "http.server"
        )

        # Going forward, the sanitization of data will need to happen within the `before_send_span` hooks
        # See https://sentry.slack.com/archives/C09RR0KD2N7/p1776951331206129?thread_ts=1776951227.440659&cid=C09RR0KD2N7
        parsed_form_attribute = json.loads(
            server_span["attributes"][SPANDATA.HTTP_REQUEST_BODY_DATA]
        )
        assert parsed_form_attribute.keys() == PARSED_FORM.keys()
        assert parsed_form_attribute["username"] == PARSED_FORM["username"]
        assert parsed_form_attribute["password"] == "hello123"
        assert parsed_form_attribute["photo"] == "[Unparsable]"
    else:
        events = capture_events()

        client.post(
            "/body/form",
            data=BODY_FORM.encode("utf-8"),
            headers={
                "content-type": "multipart/form-data; boundary=fd721ef49ea403a6",
            },
        )

        (event, transaction_event) = events
        assert event["request"]["data"].keys() == PARSED_FORM.keys()
        assert event["request"]["data"]["username"] == PARSED_FORM["username"]
        assert event["request"]["data"]["password"] == "[Filtered]"
        assert event["request"]["data"]["photo"] == ""
        assert event["_meta"]["request"]["data"]["photo"] == {
            "": {"rem": [["!raw", "x"]]}
        }

        assert transaction_event["request"]["data"].keys() == PARSED_FORM.keys()
        assert (
            transaction_event["request"]["data"]["username"] == PARSED_FORM["username"]
        )
        assert transaction_event["request"]["data"]["password"] == "[Filtered]"
        assert transaction_event["request"]["data"]["photo"] == ""
        assert transaction_event["_meta"]["request"]["data"]["photo"] == {
            "": {"rem": [["!raw", "x"]]}
        }


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_request_body_too_big(
    sentry_init, capture_events, capture_items, span_streaming
):
    sentry_init(
        traces_sample_rate=1.0,
        send_default_pii=True,
        integrations=[StarletteIntegration()],
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    starlette_app = starlette_app_factory()
    client = TestClient(starlette_app)

    if span_streaming:
        items = capture_items("event", "span")

        client.post(
            "/body/form",
            data=BODY_FORM.encode("utf-8"),
            headers={
                "content-type": "multipart/form-data; boundary=fd721ef49ea403a6",
                "cookie": "yummy_cookie=choco; tasty_cookie=strawberry",
            },
        )

        (event,) = (item.payload for item in items if item.type == "event")
        assert event["request"]["cookies"] == {
            "tasty_cookie": "strawberry",
            "yummy_cookie": "choco",
        }
        # Because request is too big only the AnnotatedValue is extracted.
        assert event["_meta"]["request"]["data"] == {"": {"rem": [["!config", "x"]]}}

        sentry_sdk.flush()
        spans = [item.payload for item in items if item.type == "span"]
        server_span = next(
            span for span in spans if span["attributes"]["sentry.op"] == "http.server"
        )

        # Because request is too big only the AnnotatedValue is extracted.
        assert (
            server_span["attributes"][SPANDATA.HTTP_REQUEST_BODY_DATA]
            == "[Exceeds maximum size]"
        )
    else:
        events = capture_events()

        client.post(
            "/body/form",
            data=BODY_FORM.encode("utf-8"),
            headers={
                "content-type": "multipart/form-data; boundary=fd721ef49ea403a6",
                "cookie": "yummy_cookie=choco; tasty_cookie=strawberry",
            },
        )

        (event, transaction_event) = events
        assert event["request"]["cookies"] == {
            "tasty_cookie": "strawberry",
            "yummy_cookie": "choco",
        }
        # Because request is too big only the AnnotatedValue is extracted.
        assert event["_meta"]["request"]["data"] == {"": {"rem": [["!config", "x"]]}}

        assert transaction_event["request"]["cookies"] == {
            "tasty_cookie": "strawberry",
            "yummy_cookie": "choco",
        }
        # Because request is too big only the AnnotatedValue is extracted.
        assert transaction_event["_meta"]["request"]["data"] == {
            "": {"rem": [["!config", "x"]]}
        }


@pytest.mark.asyncio
async def test_request_info_no_pii(sentry_init, capture_events):
    sentry_init(
        traces_sample_rate=1.0,
        send_default_pii=False,
        integrations=[StarletteIntegration()],
    )

    starlette_app = starlette_app_factory()
    events = capture_events()

    client = TestClient(starlette_app)
    client.post(
        "/body/json",
        json=BODY_JSON,
        headers={
            "cookie": "yummy_cookie=choco; tasty_cookie=strawberry",
        },
    )

    (event, transaction_event) = events
    assert "cookies" not in event["request"]
    assert event["request"]["data"] == BODY_JSON

    assert "cookies" not in transaction_event["request"]
    assert transaction_event["request"]["data"] == BODY_JSON


@pytest.mark.parametrize(
    "url,transaction_style,expected_transaction,expected_source",
    [
        (
            "/message",
            "url",
            "/message",
            "route",
        ),
        (
            "/message",
            "endpoint",
            "tests.integrations.starlette.test_starlette.starlette_app_factory.<locals>._message",
            "component",
        ),
        (
            "/message/123456",
            "url",
            "/message/{message_id}",
            "route",
        ),
        (
            "/message/123456",
            "endpoint",
            "tests.integrations.starlette.test_starlette.starlette_app_factory.<locals>._message_with_id",
            "component",
        ),
    ],
)
def test_transaction_style(
    sentry_init,
    capture_events,
    url,
    transaction_style,
    expected_transaction,
    expected_source,
):
    sentry_init(
        integrations=[StarletteIntegration(transaction_style=transaction_style)],
    )
    starlette_app = starlette_app_factory()

    events = capture_events()

    client = TestClient(starlette_app)
    client.get(url)

    (event,) = events
    assert event["transaction"] == expected_transaction
    assert event["transaction_info"] == {"source": expected_source}


@pytest.mark.parametrize(
    "test_url,expected_error,expected_message",
    [
        ("/some_url", ZeroDivisionError, "division by zero"),
        ("/custom_error", Exception, "Too Hot"),
    ],
)
def test_catch_exceptions(
    sentry_init,
    capture_exceptions,
    capture_events,
    test_url,
    expected_error,
    expected_message,
):
    sentry_init(integrations=[StarletteIntegration()])
    starlette_app = starlette_app_factory()
    exceptions = capture_exceptions()
    events = capture_events()

    client = TestClient(starlette_app)
    try:
        client.get(test_url)
    except Exception:
        pass

    (exc,) = exceptions
    assert isinstance(exc, expected_error)
    assert str(exc) == expected_message

    (event,) = events
    assert event["exception"]["values"][0]["mechanism"]["type"] == "starlette"


def test_user_information_error(sentry_init, capture_events):
    sentry_init(
        send_default_pii=True,
        integrations=[StarletteIntegration()],
    )
    starlette_app = starlette_app_factory(
        middleware=[Middleware(AuthenticationMiddleware, backend=BasicAuthBackend())]
    )
    events = capture_events()

    client = TestClient(starlette_app, raise_server_exceptions=False)
    try:
        client.get("/custom_error", auth=("Gabriela", "hello123"))
    except Exception:
        pass

    (event,) = events
    user = event.get("user", None)
    assert user
    assert "username" in user
    assert user["username"] == "Gabriela"


def test_user_information_error_no_pii(sentry_init, capture_events):
    sentry_init(
        send_default_pii=False,
        integrations=[StarletteIntegration()],
    )
    starlette_app = starlette_app_factory(
        middleware=[Middleware(AuthenticationMiddleware, backend=BasicAuthBackend())]
    )
    events = capture_events()

    client = TestClient(starlette_app, raise_server_exceptions=False)
    try:
        client.get("/custom_error", auth=("Gabriela", "hello123"))
    except Exception:
        pass

    (event,) = events
    assert "user" not in event


def test_user_information_transaction(sentry_init, capture_events):
    sentry_init(
        traces_sample_rate=1.0,
        send_default_pii=True,
        integrations=[StarletteIntegration()],
    )
    starlette_app = starlette_app_factory(
        middleware=[Middleware(AuthenticationMiddleware, backend=BasicAuthBackend())]
    )
    events = capture_events()

    client = TestClient(starlette_app, raise_server_exceptions=False)
    client.get("/message", auth=("Gabriela", "hello123"))

    (_, transaction_event) = events
    user = transaction_event.get("user", None)
    assert user
    assert "username" in user
    assert user["username"] == "Gabriela"


def test_user_information_transaction_no_pii(sentry_init, capture_events):
    sentry_init(
        traces_sample_rate=1.0,
        send_default_pii=False,
        integrations=[StarletteIntegration()],
    )
    starlette_app = starlette_app_factory(
        middleware=[Middleware(AuthenticationMiddleware, backend=BasicAuthBackend())]
    )
    events = capture_events()

    client = TestClient(starlette_app, raise_server_exceptions=False)
    client.get("/message", auth=("Gabriela", "hello123"))

    (_, transaction_event) = events
    assert "user" not in transaction_event


def test_user_information_does_not_clobber_app_set_user(sentry_init, capture_events):
    """
    Regression test for PY-2596/GH-6756: the AuthenticationMiddleware patch must add the
    request-derived user *before* the request handler runs, so that a user the app
    sets during the request (via ``sentry_sdk.set_user``) is preserved rather than
    overwritten. Starlette's ``SimpleUser`` only carries a username, so an app-set
    ``id``/``email`` would otherwise be lost.
    """
    sentry_init(
        traces_sample_rate=1.0,
        send_default_pii=True,
        integrations=[StarletteIntegration()],
    )

    async def _set_user(request):
        sentry_sdk.set_user(
            {
                "id": "user_42",
                "email": "ada@beans.com",
                "username": "Ada",
            }
        )
        return starlette.responses.JSONResponse({"status": "ok"})

    app = starlette.applications.Starlette(
        routes=[starlette.routing.Route("/set_user", _set_user)],
        middleware=[Middleware(AuthenticationMiddleware, backend=BasicAuthBackend())],
    )

    events = capture_events()

    client = TestClient(app, raise_server_exceptions=False)
    client.get("/set_user", auth=("Ada", "hello123"))

    (transaction_event,) = events
    user = transaction_event.get("user", None)
    assert user
    assert user["username"] == "Ada"
    assert user["id"] == "user_42"
    assert user["email"] == "ada@beans.com"


@pytest.mark.parametrize("span_streaming", [True, False])
def test_middleware_spans(sentry_init, capture_events, capture_items, span_streaming):
    sentry_init(
        traces_sample_rate=1.0,
        integrations=[StarletteIntegration(middleware_spans=True)],
        auto_enabling_integrations=False,  # disable because httpx will enable otherwise, leading to the segment span being an `http.client` sentry.op (the TestClient initiating the request), rather than the more realistic `http.server`.
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )
    starlette_app = starlette_app_factory(
        middleware=[Middleware(AuthenticationMiddleware, backend=BasicAuthBackend())]
    )

    if span_streaming:
        items = capture_items("span")
    else:
        events = capture_events()

    client = TestClient(starlette_app, raise_server_exceptions=False)
    try:
        client.get("/message", auth=("Gabriela", "hello123"))
    except Exception:
        pass

    expected_middleware_spans = [
        "ServerErrorMiddleware",
        "AuthenticationMiddleware",
        "ExceptionMiddleware",
        "AuthenticationMiddleware",  # 'op': 'middleware.starlette.send'
        "ServerErrorMiddleware",  # 'op': 'middleware.starlette.send'
        "AuthenticationMiddleware",  # 'op': 'middleware.starlette.send'
        "ServerErrorMiddleware",  # 'op': 'middleware.starlette.send'
    ]

    if span_streaming:
        sentry_sdk.flush()

        segment = items.pop().payload
        middleware_spans = [item.payload for item in items]

        # In span-first, the `middleware.starlette.send` ops appear first,
        # so the list needs to be reversed for the assertions below
        middleware_spans.reverse()

        assert len(middleware_spans) == len(expected_middleware_spans)

        assert segment["is_segment"] is True
        assert segment["attributes"]["sentry.op"] == "http.server"

        idx = 0
        for idx, span in enumerate(middleware_spans):
            assert (
                span["attributes"]["middleware.name"] == expected_middleware_spans[idx]
            )
    else:
        (_, transaction_event) = events

        assert len(transaction_event["spans"]) == len(expected_middleware_spans)

        idx = 0
        for span in transaction_event["spans"]:
            if span["op"].startswith("middleware.starlette"):
                assert (
                    span["tags"]["starlette.middleware_name"]
                    == expected_middleware_spans[idx]
                )
                idx += 1


@pytest.mark.parametrize("span_streaming", [True, False])
def test_middleware_spans_disabled(
    sentry_init, capture_events, capture_items, span_streaming
):
    sentry_init(
        traces_sample_rate=1.0,
        integrations=[StarletteIntegration(middleware_spans=False)],
        auto_enabling_integrations=False,  # disable because httpx will enable otherwise, leading to the segment span being an `http.client` sentry.op (the TestClient initiating the request), rather than the more realistic `http.server`.
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )
    starlette_app = starlette_app_factory(
        middleware=[Middleware(AuthenticationMiddleware, backend=BasicAuthBackend())]
    )

    if span_streaming:
        items = capture_items("span")
    else:
        events = capture_events()

    client = TestClient(starlette_app, raise_server_exceptions=False)
    try:
        client.get("/message", auth=("Gabriela", "hello123"))
    except Exception:
        pass

    if span_streaming:
        sentry_sdk.flush()

        segment = items.pop().payload
        middleware_spans = [item.payload for item in items]

        assert len(middleware_spans) == 0

        assert segment["is_segment"] is True
        assert segment["attributes"]["sentry.op"] == "http.server"
    else:
        (_, transaction_event) = events
        assert len(transaction_event["spans"]) == 0


@pytest.mark.parametrize("span_streaming", [True, False])
def test_middleware_callback_spans(
    sentry_init, capture_events, capture_items, span_streaming
):
    sentry_init(
        traces_sample_rate=1.0,
        integrations=[StarletteIntegration(middleware_spans=True)],
        auto_enabling_integrations=False,  # disable because httpx will enable otherwise, leading to the segment span being an `http.client` sentry.op (the TestClient initiating the request), rather than the more realistic `http.server`.
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )
    starlette_app = starlette_app_factory(middleware=[Middleware(SampleMiddleware)])

    if span_streaming:
        items = capture_items("span")
    else:
        events = capture_events()

    client = TestClient(starlette_app, raise_server_exceptions=False)
    try:
        client.get("/message", auth=("Gabriela", "hello123"))
    except Exception:
        pass

    expected = [
        {
            "op": "middleware.starlette",
            "description": "ServerErrorMiddleware",
            "tags": {"starlette.middleware_name": "ServerErrorMiddleware"},
        },
        {
            "op": "middleware.starlette",
            "description": "SampleMiddleware",
            "tags": {"starlette.middleware_name": "SampleMiddleware"},
        },
        {
            "op": "middleware.starlette",
            "description": "ExceptionMiddleware",
            "tags": {"starlette.middleware_name": "ExceptionMiddleware"},
        },
        {
            "op": "middleware.starlette.send",
            "description": "SampleMiddleware.__call__.<locals>.do_stuff",
            "tags": {"starlette.middleware_name": "ExceptionMiddleware"},
        },
        {
            "op": "middleware.starlette.send",
            "description": "ServerErrorMiddleware.__call__.<locals>._send",
            "tags": {"starlette.middleware_name": "SampleMiddleware"},
        },
        {
            "op": "middleware.starlette.send",
            "description": "SentryAsgiMiddleware._run_app.<locals>._sentry_wrapped_send",
            "tags": {"starlette.middleware_name": "ServerErrorMiddleware"},
        },
        {
            "op": "middleware.starlette.send",
            "description": "SampleMiddleware.__call__.<locals>.do_stuff",
            "tags": {"starlette.middleware_name": "ExceptionMiddleware"},
        },
        {
            "op": "middleware.starlette.send",
            "description": "ServerErrorMiddleware.__call__.<locals>._send",
            "tags": {"starlette.middleware_name": "SampleMiddleware"},
        },
        {
            "op": "middleware.starlette.send",
            "description": "SentryAsgiMiddleware._run_app.<locals>._sentry_wrapped_send",
            "tags": {"starlette.middleware_name": "ServerErrorMiddleware"},
        },
    ]

    if span_streaming:
        sentry_sdk.flush()

        segment = items.pop().payload
        middleware_spans = [item.payload for item in items]

        # In span-first, the `middleware.starlette.send` ops appear first,
        # so the list needs to be reversed for the assertions below
        middleware_spans.reverse()

        assert len(middleware_spans) == len(expected)

        assert segment["is_segment"] is True
        assert segment["attributes"]["sentry.op"] == "http.server"

        for span, exp in zip(middleware_spans, expected):
            assert span["attributes"]["sentry.op"] == exp["op"]
            assert span["name"] == exp["description"]
            assert (
                span["attributes"]["middleware.name"]
                == exp["tags"]["starlette.middleware_name"]
            )
    else:
        (_, transaction_event) = events

        idx = 0
        for span in transaction_event["spans"]:
            assert span["op"] == expected[idx]["op"]
            assert span["description"] == expected[idx]["description"]
            assert span["tags"] == expected[idx]["tags"]
            idx += 1


def test_middleware_receive_send(sentry_init, capture_events):
    sentry_init(
        traces_sample_rate=1.0,
        integrations=[StarletteIntegration()],
    )
    starlette_app = starlette_app_factory(
        middleware=[Middleware(SampleReceiveSendMiddleware)]
    )

    client = TestClient(starlette_app, raise_server_exceptions=False)
    try:
        # NOTE: the assert statements checking
        # for correct behaviour are in `SampleReceiveSendMiddleware`!
        client.get("/message", auth=("Gabriela", "hello123"))
    except Exception:
        pass


def test_middleware_partial_receive_send(sentry_init, capture_events):
    sentry_init(
        traces_sample_rate=1.0,
        integrations=[StarletteIntegration()],
    )
    starlette_app = starlette_app_factory(
        middleware=[Middleware(SamplePartialReceiveSendMiddleware)]
    )
    events = capture_events()

    client = TestClient(starlette_app, raise_server_exceptions=False)
    try:
        client.get("/message", auth=("Gabriela", "hello123"))
    except Exception:
        pass

    (_, transaction_event) = events

    expected = [
        {
            "op": "middleware.starlette",
            "description": "ServerErrorMiddleware",
            "tags": {"starlette.middleware_name": "ServerErrorMiddleware"},
        },
        {
            "op": "middleware.starlette",
            "description": "SamplePartialReceiveSendMiddleware",
            "tags": {"starlette.middleware_name": "SamplePartialReceiveSendMiddleware"},
        },
        {
            "op": "middleware.starlette.receive",
            "description": (
                "_ASGIAdapter.send.<locals>.receive"
                if STARLETTE_VERSION < (0, 21)
                else "_TestClientTransport.handle_request.<locals>.receive"
            ),
            "tags": {"starlette.middleware_name": "ServerErrorMiddleware"},
        },
        {
            "op": "middleware.starlette.send",
            "description": "ServerErrorMiddleware.__call__.<locals>._send",
            "tags": {"starlette.middleware_name": "SamplePartialReceiveSendMiddleware"},
        },
        {
            "op": "middleware.starlette.send",
            "description": "SentryAsgiMiddleware._run_app.<locals>._sentry_wrapped_send",
            "tags": {"starlette.middleware_name": "ServerErrorMiddleware"},
        },
        {
            "op": "middleware.starlette",
            "description": "ExceptionMiddleware",
            "tags": {"starlette.middleware_name": "ExceptionMiddleware"},
        },
        {
            "op": "middleware.starlette.send",
            "description": "functools.partial(<function SamplePartialReceiveSendMiddleware.__call__.<locals>.my_send at ",
            "tags": {"starlette.middleware_name": "ExceptionMiddleware"},
        },
        {
            "op": "middleware.starlette.send",
            "description": "functools.partial(<function SamplePartialReceiveSendMiddleware.__call__.<locals>.my_send at ",
            "tags": {"starlette.middleware_name": "ExceptionMiddleware"},
        },
    ]

    idx = 0
    for span in transaction_event["spans"]:
        assert span["op"] == expected[idx]["op"]
        assert span["description"].startswith(expected[idx]["description"])
        assert span["tags"] == expected[idx]["tags"]
        idx += 1


@pytest.mark.skipif(
    STARLETTE_VERSION < (0, 35),
    reason="Positional args for middleware have been introduced in Starlette >= 0.35",
)
def test_middleware_positional_args(sentry_init):
    sentry_init(
        traces_sample_rate=1.0,
        integrations=[StarletteIntegration()],
    )
    _ = starlette_app_factory(middleware=[Middleware(SampleMiddlewareWithArgs, "bla")])

    # Only creating the App with an Middleware with args
    # should not raise an error
    # So as long as test passes, we are good


def test_legacy_setup(
    sentry_init,
    capture_events,
):
    # Check that behaviour does not change
    # if the user just adds the new Integration
    # and forgets to remove SentryAsgiMiddleware
    sentry_init()
    app = starlette_app_factory()
    asgi_app = SentryAsgiMiddleware(app)

    events = capture_events()

    client = TestClient(asgi_app)
    client.get("/message/123456")

    (event,) = events
    assert event["transaction"] == "/message/{message_id}"


@pytest.mark.parametrize("endpoint", ["/sync/thread_ids", "/async/thread_ids"])
@mock.patch("sentry_sdk.profiler.transaction_profiler.PROFILE_MINIMUM_SAMPLES", 0)
def test_active_thread_id(sentry_init, capture_envelopes, teardown_profiling, endpoint):
    sentry_init(
        traces_sample_rate=1.0,
        profiles_sample_rate=1.0,
    )
    app = starlette_app_factory()
    asgi_app = SentryAsgiMiddleware(app)

    envelopes = capture_envelopes()

    client = TestClient(asgi_app)
    response = client.get(endpoint)
    assert response.status_code == 200

    data = json.loads(response.content)

    envelopes = [envelope for envelope in envelopes]
    assert len(envelopes) == 1

    profiles = [item for item in envelopes[0].items if item.type == "profile"]
    assert len(profiles) == 1

    for item in profiles:
        transactions = item.payload.json["transactions"]
        assert len(transactions) == 1
        assert str(data["active"]) == transactions[0]["active_thread_id"]

    transactions = [item for item in envelopes[0].items if item.type == "transaction"]
    assert len(transactions) == 1

    for item in transactions:
        transaction = item.payload.json
        trace_context = transaction["contexts"]["trace"]
        assert str(data["active"]) == trace_context["data"]["thread.id"]


@pytest.mark.parametrize("endpoint", ["/sync/thread_ids", "/async/thread_ids"])
def test_active_thread_id_span_streaming(sentry_init, capture_items, endpoint):
    sentry_init(
        auto_enabling_integrations=False,  # avoid legacy spans from auto-enabled integrations leaking into streaming mode
        integrations=[StarletteIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )
    app = starlette_app_factory()

    items = capture_items("span")

    client = TestClient(app)
    response = client.get(endpoint)
    assert response.status_code == 200

    data = json.loads(response.content)

    sentry_sdk.flush()

    segments = [item.payload for item in items if item.payload.get("is_segment")]
    assert len(segments) == 1
    assert str(data["active"]) == segments[0]["attributes"]["thread.id"]


def test_original_request_not_scrubbed(sentry_init, capture_events):
    sentry_init(integrations=[StarletteIntegration()])

    events = capture_events()

    async def _error(request):
        logging.critical("Oh no!")
        assert request.headers["Authorization"] == "Bearer ohno"
        assert request.headers["Proxy-Authorization"] == "Basic ohno"
        assert await request.json() == {"password": "ohno"}
        return starlette.responses.JSONResponse({"status": "Oh no!"})

    app = starlette.applications.Starlette(
        routes=[
            starlette.routing.Route("/error", _error, methods=["POST"]),
        ],
    )

    client = TestClient(app)
    client.post(
        "/error",
        json={"password": "ohno"},
        headers={
            "Authorization": "Bearer ohno",
            "Proxy-Authorization": "Basic ohno",
        },
    )

    event = events[0]
    assert event["request"]["data"] == {"password": "[Filtered]"}
    assert event["request"]["headers"]["authorization"] == "[Filtered]"
    assert event["request"]["headers"]["proxy-authorization"] == "[Filtered]"


@pytest.mark.skipif(STARLETTE_VERSION < (0, 24), reason="Requires Starlette >= 0.24")
def test_template_tracing_meta(sentry_init, capture_events):
    sentry_init(
        auto_enabling_integrations=False,  # Make sure that httpx integration is not added, because it adds tracing information to the starlette test clients request.
        integrations=[StarletteIntegration()],
    )
    events = capture_events()

    app = starlette_app_factory()

    client = TestClient(app)
    response = client.get("/render_template")
    assert response.status_code == 200

    rendered_meta = response.text
    traceparent, baggage = events[0]["message"].split("\n")
    assert traceparent != ""
    assert baggage != ""

    match = re.match(
        r'^<meta name="sentry-trace" content="([^\"]*)"><meta name="baggage" content="([^\"]*)">',
        rendered_meta,
    )
    assert match is not None
    assert match.group(1) == traceparent

    rendered_baggage = match.group(2)
    assert rendered_baggage == baggage


@pytest.mark.parametrize(
    "request_url,transaction_style,expected_transaction_name,expected_transaction_source",
    [
        (
            "/message/123456",
            "endpoint",
            "tests.integrations.starlette.test_starlette.starlette_app_factory.<locals>._message_with_id",
            "component",
        ),
        (
            "/message/123456",
            "url",
            "/message/{message_id}",
            "route",
        ),
    ],
)
def test_transaction_name(
    sentry_init,
    request_url,
    transaction_style,
    expected_transaction_name,
    expected_transaction_source,
    capture_envelopes,
):
    """
    Tests that the transaction name is something meaningful.
    """
    sentry_init(
        auto_enabling_integrations=False,  # Make sure that httpx integration is not added, because it adds tracing information to the starlette test clients request.
        integrations=[StarletteIntegration(transaction_style=transaction_style)],
        traces_sample_rate=1.0,
    )

    envelopes = capture_envelopes()

    app = starlette_app_factory()
    client = TestClient(app)
    client.get(request_url)

    (_, transaction_envelope) = envelopes
    transaction_event = transaction_envelope.get_transaction_event()

    assert transaction_event["transaction"] == expected_transaction_name
    assert (
        transaction_event["transaction_info"]["source"] == expected_transaction_source
    )


@pytest.mark.parametrize(
    "request_url,transaction_style,expected_transaction_name,expected_transaction_source",
    [
        (
            "/message/123456",
            "endpoint",
            "http://testserver/message/123456",
            "url",
        ),
        (
            "/message/123456",
            "url",
            "http://testserver/message/123456",
            "url",
        ),
    ],
)
def test_transaction_name_in_traces_sampler(
    sentry_init,
    request_url,
    transaction_style,
    expected_transaction_name,
    expected_transaction_source,
):
    """
    Tests that a custom traces_sampler has a meaningful transaction name.
    In this case the URL or endpoint, because we do not have the route yet.
    """

    def dummy_traces_sampler(sampling_context):
        assert (
            sampling_context["transaction_context"]["name"] == expected_transaction_name
        )
        assert (
            sampling_context["transaction_context"]["source"]
            == expected_transaction_source
        )

    sentry_init(
        auto_enabling_integrations=False,  # Make sure that httpx integration is not added, because it adds tracing information to the starlette test clients request.
        integrations=[StarletteIntegration(transaction_style=transaction_style)],
        traces_sampler=dummy_traces_sampler,
        traces_sample_rate=1.0,
    )

    app = starlette_app_factory()
    client = TestClient(app)
    client.get(request_url)


@pytest.mark.parametrize("middleware_spans", [False, True])
@pytest.mark.parametrize(
    "request_url,transaction_style,expected_transaction_name,expected_transaction_source",
    [
        (
            "/message/123456",
            "endpoint",
            "starlette.middleware.trustedhost.TrustedHostMiddleware",
            "component",
        ),
        (
            "/message/123456",
            "url",
            "http://testserver/message/123456",
            "url",
        ),
    ],
)
def test_transaction_name_in_middleware(
    sentry_init,
    middleware_spans,
    request_url,
    transaction_style,
    expected_transaction_name,
    expected_transaction_source,
    capture_envelopes,
):
    """
    Tests that the transaction name is something meaningful.
    """
    sentry_init(
        auto_enabling_integrations=False,  # Make sure that httpx integration is not added, because it adds tracing information to the starlette test clients request.
        integrations=[
            StarletteIntegration(
                transaction_style=transaction_style, middleware_spans=middleware_spans
            ),
        ],
        traces_sample_rate=1.0,
    )

    envelopes = capture_envelopes()

    middleware = [
        Middleware(
            TrustedHostMiddleware,
            allowed_hosts=["example.com", "*.example.com"],
        ),
    ]

    app = starlette_app_factory(middleware=middleware)
    client = TestClient(app)
    client.get(request_url)

    (transaction_envelope,) = envelopes
    transaction_event = transaction_envelope.get_transaction_event()

    assert transaction_event["contexts"]["response"]["status_code"] == 400
    assert transaction_event["transaction"] == expected_transaction_name
    assert (
        transaction_event["transaction_info"]["source"] == expected_transaction_source
    )


@pytest.mark.parametrize("span_streaming", [True, False])
def test_span_origin(sentry_init, capture_events, capture_items, span_streaming):
    sentry_init(
        auto_enabling_integrations=False,  # avoid httpx auto-instrumentation leaking spans
        integrations=[StarletteIntegration(middleware_spans=True)],
        traces_sample_rate=1.0,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )
    starlette_app = starlette_app_factory(
        middleware=[Middleware(AuthenticationMiddleware, backend=BasicAuthBackend())]
    )

    if span_streaming:
        items = capture_items("span")
    else:
        events = capture_events()

    client = TestClient(starlette_app, raise_server_exceptions=False)
    try:
        client.get("/message", auth=("Gabriela", "hello123"))
    except Exception:
        pass

    if span_streaming:
        sentry_sdk.flush()

        assert len(items) > 0
        for item in items:
            assert item.payload["attributes"]["sentry.origin"] == "auto.http.starlette"
    else:
        (_, event) = events

        assert event["contexts"]["trace"]["origin"] == "auto.http.starlette"
        for span in event["spans"]:
            assert span["origin"] == "auto.http.starlette"


class NonIterableContainer:
    """Wraps any container and makes it non-iterable.

    Used to test backwards compatibility with our old way of defining failed_request_status_codes, which allowed
    passing in a list of (possibly non-iterable) containers. The Python standard library does not provide any built-in
    non-iterable containers, so we have to define our own.
    """

    def __init__(self, inner):
        self.inner = inner

    def __contains__(self, item):
        return item in self.inner


parametrize_test_configurable_status_codes_deprecated = pytest.mark.parametrize(
    "failed_request_status_codes,status_code,expected_error",
    [
        (None, 500, True),
        (None, 400, False),
        ([500, 501], 500, True),
        ([500, 501], 401, False),
        ([range(400, 499)], 401, True),
        ([range(400, 499)], 500, False),
        ([range(400, 499), range(500, 599)], 300, False),
        ([range(400, 499), range(500, 599)], 403, True),
        ([range(400, 499), range(500, 599)], 503, True),
        ([range(400, 403), 500, 501], 401, True),
        ([range(400, 403), 500, 501], 405, False),
        ([range(400, 403), 500, 501], 501, True),
        ([range(400, 403), 500, 501], 503, False),
        ([], 500, False),
        ([NonIterableContainer(range(500, 600))], 500, True),
        ([NonIterableContainer(range(500, 600))], 404, False),
    ],
)
"""Test cases for configurable status codes (deprecated API).
Also used by the FastAPI tests.
"""


@parametrize_test_configurable_status_codes_deprecated
def test_configurable_status_codes_deprecated(
    sentry_init,
    capture_events,
    failed_request_status_codes,
    status_code,
    expected_error,
):
    with pytest.warns(DeprecationWarning):
        starlette_integration = StarletteIntegration(
            failed_request_status_codes=failed_request_status_codes
        )

    sentry_init(integrations=[starlette_integration])

    events = capture_events()

    async def _error(request):
        raise HTTPException(status_code)

    app = starlette.applications.Starlette(
        routes=[
            starlette.routing.Route("/error", _error, methods=["GET"]),
        ],
    )

    client = TestClient(app)
    client.get("/error")

    if expected_error:
        assert len(events) == 1
    else:
        assert not events


@pytest.mark.skipif(
    STARLETTE_VERSION < (0, 21),
    reason="Requires Starlette >= 0.21, because earlier versions do not support HTTP 'HEAD' requests",
)
def test_transaction_http_method_default(sentry_init, capture_events):
    """
    By default OPTIONS and HEAD requests do not create a transaction.
    """
    sentry_init(
        traces_sample_rate=1.0,
        integrations=[
            StarletteIntegration(),
        ],
    )
    events = capture_events()

    starlette_app = starlette_app_factory()

    client = TestClient(starlette_app)
    client.get("/nomessage")
    client.options("/nomessage")
    client.head("/nomessage")

    assert len(events) == 1

    (event,) = events

    assert event["request"]["method"] == "GET"


@pytest.mark.parametrize("span_streaming", [True, False])
def test_request_url(sentry_init, capture_events, capture_items, span_streaming):
    sentry_init(
        traces_sample_rate=1.0,
        send_default_pii=True,
        integrations=[
            StarletteIntegration(),
        ],
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    starlette_app = starlette_app_factory()

    client = TestClient(starlette_app)

    if span_streaming:
        items = capture_items("span")

        client.get("/root/nomessage")
        sentry_sdk.flush()
        spans = [item.payload for item in items]

        (server_span,) = (
            span
            for span in spans
            if span["attributes"].get("sentry.op") == "http.server"
        )
        assert server_span["attributes"][SPANDATA.URL_FULL] == (
            "http://testserver/root/nomessage"
        )
        assert server_span["attributes"][SPANDATA.URL_PATH] == "/root/nomessage"
    else:
        events = capture_events()

        client.get("/root/nomessage")

        (event,) = events
        assert event["request"]["url"] == "http://testserver/root/nomessage"


@pytest.mark.skipif(
    STARLETTE_VERSION < (0, 21),
    reason="Requires Starlette >= 0.21, because earlier versions do not support HTTP 'HEAD' requests",
)
def test_transaction_http_method_custom(sentry_init, capture_events):
    sentry_init(
        traces_sample_rate=1.0,
        integrations=[
            StarletteIntegration(
                http_methods_to_capture=(
                    "OPTIONS",
                    "head",
                ),  # capitalization does not matter
            ),
        ],
        debug=True,
    )
    events = capture_events()

    starlette_app = starlette_app_factory()

    client = TestClient(starlette_app)
    client.get("/nomessage")
    client.options("/nomessage")
    client.head("/nomessage")

    assert len(events) == 2

    (event1, event2) = events

    assert event1["request"]["method"] == "OPTIONS"
    assert event2["request"]["method"] == "HEAD"


@parametrize_test_configurable_status_codes
def test_configurable_status_codes(
    sentry_init,
    capture_events,
    failed_request_status_codes,
    status_code,
    expected_error,
):
    integration_kwargs = {}
    if failed_request_status_codes is not None:
        integration_kwargs["failed_request_status_codes"] = failed_request_status_codes

    with warnings.catch_warnings():
        warnings.simplefilter("error", DeprecationWarning)
        starlette_integration = StarletteIntegration(**integration_kwargs)

    sentry_init(integrations=[starlette_integration])

    events = capture_events()

    async def _error(_):
        raise HTTPException(status_code)

    app = starlette.applications.Starlette(
        routes=[
            starlette.routing.Route("/error", _error, methods=["GET"]),
        ],
    )

    client = TestClient(app)
    client.get("/error")

    assert len(events) == int(expected_error)


@pytest.mark.asyncio
async def test_malformed_json_request_body(sentry_init, capture_events):
    sentry_init(
        traces_sample_rate=1.0,
        send_default_pii=True,
        integrations=[StarletteIntegration()],
    )

    starlette_app = starlette_app_factory()
    events = capture_events()

    client = TestClient(starlette_app)
    client.post(
        "/body/raw",
        data="{invalid json".encode("utf-8"),
        headers={"content-type": "application/json"},
    )

    (event, transaction_event) = events
    assert event["request"]["data"] == ""
    assert transaction_event["request"]["data"] == ""



import os
from datetime import datetime
from unittest import mock

import pytest
from sqlalchemy import Column, ForeignKey, Integer, String, create_engine, text
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, sessionmaker

import sentry_sdk
from sentry_sdk import capture_message, start_transaction
from sentry_sdk.consts import SPANDATA
from sentry_sdk.integrations.sqlalchemy import SqlalchemyIntegration
from sentry_sdk.serializer import MAX_EVENT_BYTES
from sentry_sdk.tracing_utils import record_sql_queries
from sentry_sdk.utils import json_dumps


@pytest.mark.parametrize("span_streaming", [True, False])
def test_orm_queries(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[SqlalchemyIntegration()],
        _experiments={
            "record_sql_params": True,
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    Base = declarative_base()  # noqa: N806

    class Person(Base):
        __tablename__ = "person"
        id = Column(Integer, primary_key=True)
        name = Column(String(250), nullable=False)

    class Address(Base):
        __tablename__ = "address"
        id = Column(Integer, primary_key=True)
        street_name = Column(String(250))
        street_number = Column(String(250))
        post_code = Column(String(250), nullable=False)
        person_id = Column(Integer, ForeignKey("person.id"))
        person = relationship(Person)

    engine = create_engine(
        "sqlite:///:memory:", connect_args={"check_same_thread": False}
    )
    Base.metadata.create_all(engine)

    Session = sessionmaker(bind=engine)  # noqa: N806
    session = Session()

    bob = Person(name="Bob")
    session.add(bob)

    assert session.query(Person).first() == bob

    if span_streaming:
        items = capture_items("event")
        capture_message("hi")
        (event,) = (item.payload for item in items if item.type == "event")
    else:
        events = capture_events()
        capture_message("hi")
        (event,) = events

    for crumb in event["breadcrumbs"]["values"]:
        del crumb["timestamp"]

    assert event["breadcrumbs"]["values"][-2:] == [
        {
            "category": "query",
            "data": {"db.params": ["Bob"], "db.paramstyle": "qmark"},
            "message": "INSERT INTO person (name) VALUES (?)",
            "type": "default",
        },
        {
            "category": "query",
            "data": {"db.params": [1, 0], "db.paramstyle": "qmark"},
            "message": "SELECT person.id AS person_id, person.name AS person_name \n"
            "FROM person\n"
            " LIMIT ? OFFSET ?",
            "type": "default",
        },
    ]


@pytest.mark.parametrize("span_streaming", [True, False])
def test_transactions(
    sentry_init,
    capture_events,
    capture_items,
    render_span_tree,
    span_streaming,
):
    sentry_init(
        integrations=[SqlalchemyIntegration()],
        _experiments={
            "record_sql_params": True,
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
        traces_sample_rate=1.0,
    )

    Base = declarative_base()  # noqa: N806

    class Person(Base):
        __tablename__ = "person"
        id = Column(Integer, primary_key=True)
        name = Column(String(250), nullable=False)

    class Address(Base):
        __tablename__ = "address"
        id = Column(Integer, primary_key=True)
        street_name = Column(String(250))
        street_number = Column(String(250))
        post_code = Column(String(250), nullable=False)
        person_id = Column(Integer, ForeignKey("person.id"))
        person = relationship(Person)

    engine = create_engine(
        "sqlite:///:memory:", connect_args={"check_same_thread": False}
    )
    Base.metadata.create_all(engine)

    Session = sessionmaker(bind=engine)  # noqa: N806
    session = Session()

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="custom parent"):
            with session.begin_nested():
                session.query(Person).first()

            for _ in range(2):
                with pytest.raises(IntegrityError):
                    with session.begin_nested():
                        session.add(Person(id=1, name="bob"))
                        session.add(Person(id=1, name="bob"))

            with session.begin_nested():
                session.query(Person).first()

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        sqlalchemy_spans = [
            span
            for span in spans
            if span["attributes"]["sentry.origin"] == "auto.db.sqlalchemy"
        ]
        for span in sqlalchemy_spans:
            assert span["attributes"][SPANDATA.DB_SYSTEM_NAME] == "sqlite"
            assert span["attributes"][SPANDATA.DB_DRIVER_NAME] == "pysqlite"
            assert span["attributes"][SPANDATA.DB_NAMESPACE] == ":memory:"
            assert SPANDATA.SERVER_PORT not in span["attributes"]

        assert (
            render_span_tree(spans)
            == """\
- sentry.op=null: name="custom parent"
  - sentry.op="db": name="SAVEPOINT sa_savepoint_1"
  - sentry.op="db": name="SELECT person.id AS person_id, person.name AS person_name \\nFROM person\\n LIMIT ? OFFSET ?"
  - sentry.op="db": name="RELEASE SAVEPOINT sa_savepoint_1"
  - sentry.op="db": name="SAVEPOINT sa_savepoint_2"
  - sentry.op="db": name="INSERT INTO person (id, name) VALUES (?, ?)"
  - sentry.op="db": name="ROLLBACK TO SAVEPOINT sa_savepoint_2"
  - sentry.op="db": name="SAVEPOINT sa_savepoint_3"
  - sentry.op="db": name="INSERT INTO person (id, name) VALUES (?, ?)"
  - sentry.op="db": name="ROLLBACK TO SAVEPOINT sa_savepoint_3"
  - sentry.op="db": name="SAVEPOINT sa_savepoint_4"
  - sentry.op="db": name="SELECT person.id AS person_id, person.name AS person_name \\nFROM person\\n LIMIT ? OFFSET ?"
  - sentry.op="db": name="RELEASE SAVEPOINT sa_savepoint_4"\
"""
        )
    else:
        events = capture_events()
        with start_transaction(name="test_transaction", sampled=True):
            with session.begin_nested():
                session.query(Person).first()

            for _ in range(2):
                with pytest.raises(IntegrityError):
                    with session.begin_nested():
                        session.add(Person(id=1, name="bob"))
                        session.add(Person(id=1, name="bob"))

            with session.begin_nested():
                session.query(Person).first()

        (event,) = events

        for span in event["spans"]:
            assert span["data"][SPANDATA.DB_SYSTEM] == "sqlite"
            assert span["data"][SPANDATA.DB_DRIVER_NAME] == "pysqlite"
            assert span["data"][SPANDATA.DB_NAME] == ":memory:"
            assert SPANDATA.SERVER_ADDRESS not in span["data"]
            assert SPANDATA.SERVER_PORT not in span["data"]

        assert (
            render_span_tree(event["spans"], event["contexts"]["trace"])
            == """\
- op=null: description=null
  - op="db": description="SAVEPOINT sa_savepoint_1"
  - op="db": description="SELECT person.id AS person_id, person.name AS person_name \\nFROM person\\n LIMIT ? OFFSET ?"
  - op="db": description="RELEASE SAVEPOINT sa_savepoint_1"
  - op="db": description="SAVEPOINT sa_savepoint_2"
  - op="db": description="INSERT INTO person (id, name) VALUES (?, ?)"
  - op="db": description="ROLLBACK TO SAVEPOINT sa_savepoint_2"
  - op="db": description="SAVEPOINT sa_savepoint_3"
  - op="db": description="INSERT INTO person (id, name) VALUES (?, ?)"
  - op="db": description="ROLLBACK TO SAVEPOINT sa_savepoint_3"
  - op="db": description="SAVEPOINT sa_savepoint_4"
  - op="db": description="SELECT person.id AS person_id, person.name AS person_name \\nFROM person\\n LIMIT ? OFFSET ?"
  - op="db": description="RELEASE SAVEPOINT sa_savepoint_4"\
"""
        )


@pytest.mark.parametrize("span_streaming", [True, False])
def test_transactions_no_engine_url(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[SqlalchemyIntegration()],
        _experiments={
            "record_sql_params": True,
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
        traces_sample_rate=1.0,
    )

    Base = declarative_base()  # noqa: N806

    class Person(Base):
        __tablename__ = "person"
        id = Column(Integer, primary_key=True)
        name = Column(String(250), nullable=False)

    class Address(Base):
        __tablename__ = "address"
        id = Column(Integer, primary_key=True)
        street_name = Column(String(250))
        street_number = Column(String(250))
        post_code = Column(String(250), nullable=False)
        person_id = Column(Integer, ForeignKey("person.id"))
        person = relationship(Person)

    engine = create_engine(
        "sqlite:///:memory:", connect_args={"check_same_thread": False}
    )
    engine.url = None
    Base.metadata.create_all(engine)

    Session = sessionmaker(bind=engine)  # noqa: N806
    session = Session()

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="custom parent"):
            with session.begin_nested():
                session.query(Person).first()

            for _ in range(2):
                with pytest.raises(IntegrityError):
                    with session.begin_nested():
                        session.add(Person(id=1, name="bob"))
                        session.add(Person(id=1, name="bob"))

            with session.begin_nested():
                session.query(Person).first()

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        sqlalchemy_spans = [
            span
            for span in spans
            if span["attributes"]["sentry.origin"] == "auto.db.sqlalchemy"
        ]
        for span in sqlalchemy_spans:
            assert span["attributes"][SPANDATA.DB_SYSTEM_NAME] == "sqlite"
            assert span["attributes"][SPANDATA.DB_DRIVER_NAME] == "pysqlite"
            assert SPANDATA.DB_NAME not in span["attributes"]
            assert SPANDATA.SERVER_PORT not in span["attributes"]
    else:
        events = capture_events()
        with start_transaction(name="test_transaction", sampled=True):
            with session.begin_nested():
                session.query(Person).first()

            for _ in range(2):
                with pytest.raises(IntegrityError):
                    with session.begin_nested():
                        session.add(Person(id=1, name="bob"))
                        session.add(Person(id=1, name="bob"))

            with session.begin_nested():
                session.query(Person).first()

        (event,) = events
        for span in event["spans"]:
            assert span["data"][SPANDATA.DB_SYSTEM] == "sqlite"
            assert span["data"][SPANDATA.DB_DRIVER_NAME] == "pysqlite"
            assert SPANDATA.DB_NAME not in span["data"]
            assert SPANDATA.SERVER_ADDRESS not in span["data"]
            assert SPANDATA.SERVER_PORT not in span["data"]


@pytest.mark.parametrize("span_streaming", [True, False])
def test_long_sql_query_preserved(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        traces_sample_rate=1,
        integrations=[SqlalchemyIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    engine = create_engine(
        "sqlite:///:memory:", connect_args={"check_same_thread": False}
    )
    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="custom parent"):
            with engine.connect() as con:
                con.execute(
                    text(" UNION ".join("SELECT {}".format(i) for i in range(100)))
                )

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        name = spans[0]["name"]
        assert name.startswith("SELECT 0 UNION SELECT 1")
        assert name.endswith("SELECT 98 UNION SELECT 99")
    else:
        events = capture_events()
        with start_transaction(name="test"):
            with engine.connect() as con:
                con.execute(
                    text(" UNION ".join("SELECT {}".format(i) for i in range(100)))
                )

        (event,) = events
        description = event["spans"][0]["description"]
        assert description.startswith("SELECT 0 UNION SELECT 1")
        assert description.endswith("SELECT 98 UNION SELECT 99")


@pytest.mark.parametrize("max_value_length", [1024, None])
def test_large_event_not_truncated(sentry_init, capture_events, max_value_length):
    sentry_init(
        traces_sample_rate=1,
        integrations=[SqlalchemyIntegration()],
        max_value_length=max_value_length,
    )
    events = capture_events()

    long_str = "x" * (1034)

    scope = sentry_sdk.get_isolation_scope()

    @scope.add_event_processor
    def processor(event, hint):
        event["message"] = long_str
        return event

    engine = create_engine(
        "sqlite:///:memory:", connect_args={"check_same_thread": False}
    )
    with start_transaction(name="test"):
        with engine.connect() as con:
            for _ in range(1500):
                con.execute(
                    text(" UNION ".join("SELECT {}".format(i) for i in range(100)))
                )

    (event,) = events

    assert len(json_dumps(event)) > MAX_EVENT_BYTES

    # Some spans are discarded.
    assert len(event["spans"]) == 1000

    # Span descriptions are not truncated.
    description = event["spans"][0]["description"]
    assert len(description) == 1583
    assert description.startswith("SELECT 0")
    assert description.endswith("SELECT 98 UNION SELECT 99")

    description = event["spans"][999]["description"]
    assert len(description) == 1583
    assert description.startswith("SELECT 0")
    assert description.endswith("SELECT 98 UNION SELECT 99")

    if max_value_length:
        # Smoke check that truncation of other fields has not changed.
        assert len(event["message"]) == 1024

        # The _meta for other truncated fields should be there as well.
        assert event["_meta"]["message"] == {
            "": {
                "len": 1034,
                "rem": [["!limit", "x", 1021, 1024]],
            }
        }
    else:
        assert len(event["message"]) == 1034


@pytest.mark.parametrize("span_streaming", [True, False])
def test_engine_name_not_string(
    sentry_init,
    span_streaming,
):
    sentry_init(
        integrations=[SqlalchemyIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    engine = create_engine(
        "sqlite:///:memory:", connect_args={"check_same_thread": False}
    )
    engine.dialect.name = b"sqlite"

    with engine.connect() as con:
        con.execute(text("SELECT 0"))


@pytest.mark.parametrize("span_streaming", [True, False])
def test_query_source_disabled(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_options = {
        "integrations": [SqlalchemyIntegration()],
        "traces_sample_rate": 1.0,
        "enable_db_query_source": False,
        "db_query_source_threshold_ms": 0,
        "trace_lifecycle": "stream" if span_streaming else "static",
    }

    sentry_init(**sentry_options)

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            Base = declarative_base()  # noqa: N806

            class Person(Base):
                __tablename__ = "person"
                id = Column(Integer, primary_key=True)
                name = Column(String(250), nullable=False)

            engine = create_engine(
                "sqlite:///:memory:", connect_args={"check_same_thread": False}
            )
            Base.metadata.create_all(engine)

            Session = sessionmaker(bind=engine)  # noqa: N806
            session = Session()

            bob = Person(name="Bob")
            session.add(bob)

            assert session.query(Person).first() == bob

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        for span in spans:
            if span["attributes"].get("sentry.op") == "db" and span["name"].startswith(
                "SELECT person"
            ):
                attributes = span["attributes"]

                assert SPANDATA.CODE_LINE_NUMBER not in attributes
                assert SPANDATA.CODE_NAMESPACE not in attributes
                assert SPANDATA.CODE_FILE_PATH not in attributes
                assert SPANDATA.CODE_FUNCTION not in attributes
                break
        else:
            raise AssertionError("No db span found")

    else:
        events = capture_events()

        with start_transaction(name="test_transaction", sampled=True):
            Base = declarative_base()  # noqa: N806

            class Person(Base):
                __tablename__ = "person"
                id = Column(Integer, primary_key=True)
                name = Column(String(250), nullable=False)

            engine = create_engine(
                "sqlite:///:memory:", connect_args={"check_same_thread": False}
            )
            Base.metadata.create_all(engine)

            Session = sessionmaker(bind=engine)  # noqa: N806
            session = Session()

            bob = Person(name="Bob")
            session.add(bob)

            assert session.query(Person).first() == bob

        (event,) = events

        for span in event["spans"]:
            if span.get("op") == "db" and span.get("description").startswith(
                "SELECT person"
            ):
                data = span.get("data", {})

                assert SPANDATA.CODE_LINENO not in data
                assert SPANDATA.CODE_NAMESPACE not in data
                assert SPANDATA.CODE_FILEPATH not in data
                assert SPANDATA.CODE_FUNCTION not in data
                break
        else:
            raise AssertionError("No db span found")


@pytest.mark.parametrize("enable_db_query_source", [None, True])
@pytest.mark.parametrize("span_streaming", [True, False])
def test_query_source_enabled(
    sentry_init,
    capture_events,
    capture_items,
    enable_db_query_source,
    span_streaming,
):
    sentry_options = {
        "integrations": [SqlalchemyIntegration()],
        "traces_sample_rate": 1.0,
        "db_query_source_threshold_ms": 0,
        "trace_lifecycle": "stream" if span_streaming else "static",
    }

    if enable_db_query_source is not None:
        sentry_options["enable_db_query_source"] = enable_db_query_source

    sentry_init(**sentry_options)

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            Base = declarative_base()  # noqa: N806

            class Person(Base):
                __tablename__ = "person"
                id = Column(Integer, primary_key=True)
                name = Column(String(250), nullable=False)

            engine = create_engine(
                "sqlite:///:memory:", connect_args={"check_same_thread": False}
            )
            Base.metadata.create_all(engine)

            Session = sessionmaker(bind=engine)  # noqa: N806
            session = Session()

            bob = Person(name="Bob")
            session.add(bob)

            assert session.query(Person).first() == bob

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        for span in spans:
            if span["attributes"].get("sentry.op") == "db" and span["name"].startswith(
                "SELECT person"
            ):
                attributes = span["attributes"]

                assert SPANDATA.CODE_LINE_NUMBER in attributes
                assert SPANDATA.CODE_NAMESPACE in attributes
                assert SPANDATA.CODE_FILE_PATH in attributes
                assert SPANDATA.CODE_FUNCTION in attributes
                break
        else:
            raise AssertionError("No db span found")
    else:
        events = capture_events()

        with start_transaction(name="test_transaction", sampled=True):
            Base = declarative_base()  # noqa: N806

            class Person(Base):
                __tablename__ = "person"
                id = Column(Integer, primary_key=True)
                name = Column(String(250), nullable=False)

            engine = create_engine(
                "sqlite:///:memory:", connect_args={"check_same_thread": False}
            )
            Base.metadata.create_all(engine)

            Session = sessionmaker(bind=engine)  # noqa: N806
            session = Session()

            bob = Person(name="Bob")
            session.add(bob)

            assert session.query(Person).first() == bob

        (event,) = events

        for span in event["spans"]:
            if span.get("op") == "db" and span.get("description").startswith(
                "SELECT person"
            ):
                data = span.get("data", {})

                assert SPANDATA.CODE_LINENO in data
                assert SPANDATA.CODE_NAMESPACE in data
                assert SPANDATA.CODE_FILEPATH in data
                assert SPANDATA.CODE_FUNCTION in data
                break
        else:
            raise AssertionError("No db span found")


@pytest.mark.parametrize("span_streaming", [True, False])
def test_query_source(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[SqlalchemyIntegration()],
        traces_sample_rate=1.0,
        enable_db_query_source=True,
        db_query_source_threshold_ms=0,
        trace_lifecycle="stream" if span_streaming else "static",
    )
    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            Base = declarative_base()  # noqa: N806

            class Person(Base):
                __tablename__ = "person"
                id = Column(Integer, primary_key=True)
                name = Column(String(250), nullable=False)

            engine = create_engine(
                "sqlite:///:memory:", connect_args={"check_same_thread": False}
            )
            Base.metadata.create_all(engine)

            Session = sessionmaker(bind=engine)  # noqa: N806
            session = Session()

            bob = Person(name="Bob")
            session.add(bob)

            assert session.query(Person).first() == bob

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        for span in spans:
            if span["attributes"].get("sentry.op") == "db" and span["name"].startswith(
                "SELECT person"
            ):
                attributes = span["attributes"]

                assert SPANDATA.CODE_LINE_NUMBER in attributes
                assert SPANDATA.CODE_NAMESPACE in attributes
                assert SPANDATA.CODE_FILE_PATH in attributes
                assert SPANDATA.CODE_FUNCTION in attributes

                assert type(attributes.get(SPANDATA.CODE_LINE_NUMBER)) == int
                assert attributes.get(SPANDATA.CODE_LINE_NUMBER) > 0
                assert (
                    attributes.get(SPANDATA.CODE_NAMESPACE)
                    == "tests.integrations.sqlalchemy.test_sqlalchemy"
                )
                assert attributes.get(SPANDATA.CODE_FILE_PATH).endswith(
                    "tests/integrations/sqlalchemy/test_sqlalchemy.py"
                )

                is_relative_path = attributes.get(SPANDATA.CODE_FILE_PATH)[0] != os.sep
                assert is_relative_path

                assert attributes.get(SPANDATA.CODE_FUNCTION) == "test_query_source"
                break
        else:
            raise AssertionError("No db span found")
    else:
        events = capture_events()

        with start_transaction(name="test_transaction", sampled=True):
            Base = declarative_base()  # noqa: N806

            class Person(Base):
                __tablename__ = "person"
                id = Column(Integer, primary_key=True)
                name = Column(String(250), nullable=False)

            engine = create_engine(
                "sqlite:///:memory:", connect_args={"check_same_thread": False}
            )
            Base.metadata.create_all(engine)

            Session = sessionmaker(bind=engine)  # noqa: N806
            session = Session()

            bob = Person(name="Bob")
            session.add(bob)

            assert session.query(Person).first() == bob

        (event,) = events

        for span in event["spans"]:
            if span.get("op") == "db" and span.get("description").startswith(
                "SELECT person"
            ):
                data = span.get("data", {})

                assert SPANDATA.CODE_LINENO in data
                assert SPANDATA.CODE_NAMESPACE in data
                assert SPANDATA.CODE_FILEPATH in data
                assert SPANDATA.CODE_FUNCTION in data

                assert type(data.get(SPANDATA.CODE_LINENO)) == int
                assert data.get(SPANDATA.CODE_LINENO) > 0
                assert (
                    data.get(SPANDATA.CODE_NAMESPACE)
                    == "tests.integrations.sqlalchemy.test_sqlalchemy"
                )
                assert data.get(SPANDATA.CODE_FILEPATH).endswith(
                    "tests/integrations/sqlalchemy/test_sqlalchemy.py"
                )

                is_relative_path = data.get(SPANDATA.CODE_FILEPATH)[0] != os.sep
                assert is_relative_path

                assert data.get(SPANDATA.CODE_FUNCTION) == "test_query_source"
                break
        else:
            raise AssertionError("No db span found")


@pytest.mark.parametrize("span_streaming", [True, False])
def test_query_source_with_module_in_search_path(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    """
    Test that query source is relative to the path of the module it ran in
    """
    sentry_init(
        integrations=[SqlalchemyIntegration()],
        traces_sample_rate=1.0,
        enable_db_query_source=True,
        db_query_source_threshold_ms=0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    from sqlalchemy_helpers.helpers import (
        add_model_to_session,
        query_first_model_from_session,
    )

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="custom parent"):
            Base = declarative_base()  # noqa: N806

            class Person(Base):
                __tablename__ = "person"
                id = Column(Integer, primary_key=True)
                name = Column(String(250), nullable=False)

            engine = create_engine(
                "sqlite:///:memory:", connect_args={"check_same_thread": False}
            )
            Base.metadata.create_all(engine)

            Session = sessionmaker(bind=engine)  # noqa: N806
            session = Session()

            bob = Person(name="Bob")

            add_model_to_session(bob, session)

            assert query_first_model_from_session(Person, session) == bob

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        for span in spans:
            if span["attributes"].get("sentry.op") == "db" and span["name"].startswith(
                "SELECT person"
            ):
                attributes = span["attributes"]

                assert SPANDATA.CODE_LINE_NUMBER in attributes
                assert SPANDATA.CODE_NAMESPACE in attributes
                assert SPANDATA.CODE_FILE_PATH in attributes
                assert SPANDATA.CODE_FUNCTION in attributes

                assert type(attributes.get(SPANDATA.CODE_LINE_NUMBER)) == int
                assert attributes.get(SPANDATA.CODE_LINE_NUMBER) > 0
                assert (
                    attributes.get(SPANDATA.CODE_NAMESPACE)
                    == "sqlalchemy_helpers.helpers"
                )
                assert (
                    attributes.get(SPANDATA.CODE_FILE_PATH)
                    == "sqlalchemy_helpers/helpers.py"
                )

                is_relative_path = attributes.get(SPANDATA.CODE_FILE_PATH)[0] != os.sep
                assert is_relative_path

                assert (
                    attributes.get(SPANDATA.CODE_FUNCTION)
                    == "query_first_model_from_session"
                )
                break
        else:
            raise AssertionError("No db span found")
    else:
        events = capture_events()
        with start_transaction(name="test_transaction", sampled=True):
            Base = declarative_base()  # noqa: N806

            class Person(Base):
                __tablename__ = "person"
                id = Column(Integer, primary_key=True)
                name = Column(String(250), nullable=False)

            engine = create_engine(
                "sqlite:///:memory:", connect_args={"check_same_thread": False}
            )
            Base.metadata.create_all(engine)

            Session = sessionmaker(bind=engine)  # noqa: N806
            session = Session()

            bob = Person(name="Bob")

            add_model_to_session(bob, session)

            assert query_first_model_from_session(Person, session) == bob

        (event,) = events

        for span in event["spans"]:
            if span.get("op") == "db" and span.get("description").startswith(
                "SELECT person"
            ):
                data = span.get("data", {})

                assert SPANDATA.CODE_LINENO in data
                assert SPANDATA.CODE_NAMESPACE in data
                assert SPANDATA.CODE_FILEPATH in data
                assert SPANDATA.CODE_FUNCTION in data

                assert type(data.get(SPANDATA.CODE_LINENO)) == int
                assert data.get(SPANDATA.CODE_LINENO) > 0
                assert data.get(SPANDATA.CODE_NAMESPACE) == "sqlalchemy_helpers.helpers"
                assert (
                    data.get(SPANDATA.CODE_FILEPATH) == "sqlalchemy_helpers/helpers.py"
                )

                is_relative_path = data.get(SPANDATA.CODE_FILEPATH)[0] != os.sep
                assert is_relative_path

                assert (
                    data.get(SPANDATA.CODE_FUNCTION) == "query_first_model_from_session"
                )
                break
        else:
            raise AssertionError("No db span found")


@pytest.mark.parametrize("span_streaming", [True, False])
def test_no_query_source_if_duration_too_short(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[SqlalchemyIntegration()],
        traces_sample_rate=1.0,
        enable_db_query_source=True,
        db_query_source_threshold_ms=100,
        trace_lifecycle="stream" if span_streaming else "static",
    )
    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            Base = declarative_base()  # noqa: N806

            class Person(Base):
                __tablename__ = "person"
                id = Column(Integer, primary_key=True)
                name = Column(String(250), nullable=False)

            engine = create_engine(
                "sqlite:///:memory:", connect_args={"check_same_thread": False}
            )
            Base.metadata.create_all(engine)

            Session = sessionmaker(bind=engine)  # noqa: N806
            session = Session()

            bob = Person(name="Bob")
            session.add(bob)

            class fake_record_sql_queries:  # noqa: N801
                def __init__(self, *args, **kwargs):
                    self._ctx_mgr = record_sql_queries(*args, **kwargs)

                def __enter__(self):
                    self.span = self._ctx_mgr.__enter__()
                    self.span._start_timestamp = datetime(2024, 1, 1, microsecond=0)
                    self.span._end_timestamp = datetime(2024, 1, 1, microsecond=99999)
                    return self.span

                def __exit__(self, type, value, traceback):
                    self.span._end_timestamp = None
                    self._ctx_mgr.__exit__(type, value, traceback)

            with mock.patch(
                "sentry_sdk.integrations.sqlalchemy.record_sql_queries",
                fake_record_sql_queries,
            ):
                assert session.query(Person).first() == bob

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        for span in spans:
            if span["attributes"].get("sentry.op") == "db" and span["name"].startswith(
                "SELECT person"
            ):
                attributes = span["attributes"]

                assert SPANDATA.CODE_LINE_NUMBER not in attributes
                assert SPANDATA.CODE_NAMESPACE not in attributes
                assert SPANDATA.CODE_FILE_PATH not in attributes
                assert SPANDATA.CODE_FUNCTION not in attributes
                break
        else:
            raise AssertionError("No db span found")

    else:
        events = capture_events()

        with start_transaction(name="test_transaction", sampled=True):
            Base = declarative_base()  # noqa: N806

            class Person(Base):
                __tablename__ = "person"
                id = Column(Integer, primary_key=True)
                name = Column(String(250), nullable=False)

            engine = create_engine(
                "sqlite:///:memory:", connect_args={"check_same_thread": False}
            )
            Base.metadata.create_all(engine)

            Session = sessionmaker(bind=engine)  # noqa: N806
            session = Session()

            bob = Person(name="Bob")
            session.add(bob)

            class fake_record_sql_queries:  # noqa: N801
                def __init__(self, *args, **kwargs):
                    self._ctx_mgr = record_sql_queries(*args, **kwargs)

                def __enter__(self):
                    self.span = self._ctx_mgr.__enter__()
                    self.span.start_timestamp = datetime(2024, 1, 1, microsecond=0)
                    return self.span

                def __exit__(self, type, value, traceback):
                    self._ctx_mgr.__exit__(type, value, traceback)
                    self.span.timestamp = datetime(2024, 1, 1, microsecond=99999)

            with mock.patch(
                "sentry_sdk.integrations.sqlalchemy.record_sql_queries",
                fake_record_sql_queries,
            ):
                assert session.query(Person).first() == bob

        (event,) = events

        for span in event["spans"]:
            if span.get("op") == "db" and span.get("description").startswith(
                "SELECT person"
            ):
                data = span.get("data", {})

                assert SPANDATA.CODE_LINENO not in data
                assert SPANDATA.CODE_NAMESPACE not in data
                assert SPANDATA.CODE_FILEPATH not in data
                assert SPANDATA.CODE_FUNCTION not in data
                break
        else:
            raise AssertionError("No db span found")


@pytest.mark.parametrize("span_streaming", [True, False])
def test_query_source_if_duration_over_threshold(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[SqlalchemyIntegration()],
        traces_sample_rate=1.0,
        enable_db_query_source=True,
        db_query_source_threshold_ms=100,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            Base = declarative_base()  # noqa: N806

            class Person(Base):
                __tablename__ = "person"
                id = Column(Integer, primary_key=True)
                name = Column(String(250), nullable=False)

            engine = create_engine(
                "sqlite:///:memory:", connect_args={"check_same_thread": False}
            )
            Base.metadata.create_all(engine)

            Session = sessionmaker(bind=engine)  # noqa: N806
            session = Session()

            bob = Person(name="Bob")
            session.add(bob)

            class fake_record_sql_queries:  # noqa: N801
                def __init__(self, *args, **kwargs):
                    self._ctx_mgr = record_sql_queries(*args, **kwargs)

                def __enter__(self):
                    self.span = self._ctx_mgr.__enter__()
                    self.span._start_timestamp = datetime(2024, 1, 1, microsecond=0)
                    self.span._end_timestamp = datetime(2024, 1, 1, microsecond=101000)
                    return self.span

                def __exit__(self, type, value, traceback):
                    self.span._end_timestamp = None
                    self._ctx_mgr.__exit__(type, value, traceback)

            with mock.patch(
                "sentry_sdk.integrations.sqlalchemy.record_sql_queries",
                fake_record_sql_queries,
            ):
                assert session.query(Person).first() == bob

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        for span in spans:
            if span["attributes"].get("sentry.op") == "db" and span["name"].startswith(
                "SELECT person"
            ):
                attributes = span["attributes"]

                assert SPANDATA.CODE_LINE_NUMBER in attributes
                assert SPANDATA.CODE_NAMESPACE in attributes
                assert SPANDATA.CODE_FILE_PATH in attributes
                assert SPANDATA.CODE_FUNCTION in attributes

                assert type(attributes.get(SPANDATA.CODE_LINE_NUMBER)) == int
                assert attributes.get(SPANDATA.CODE_LINE_NUMBER) > 0
                assert (
                    attributes.get(SPANDATA.CODE_NAMESPACE)
                    == "tests.integrations.sqlalchemy.test_sqlalchemy"
                )
                assert attributes.get(SPANDATA.CODE_FILE_PATH).endswith(
                    "tests/integrations/sqlalchemy/test_sqlalchemy.py"
                )

                is_relative_path = attributes.get(SPANDATA.CODE_FILE_PATH)[0] != os.sep
                assert is_relative_path

                assert (
                    attributes.get(SPANDATA.CODE_FUNCTION)
                    == "test_query_source_if_duration_over_threshold"
                )
                break
        else:
            raise AssertionError("No db span found")
    else:
        events = capture_events()

        with start_transaction(name="test_transaction", sampled=True):
            Base = declarative_base()  # noqa: N806

            class Person(Base):
                __tablename__ = "person"
                id = Column(Integer, primary_key=True)
                name = Column(String(250), nullable=False)

            engine = create_engine(
                "sqlite:///:memory:", connect_args={"check_same_thread": False}
            )
            Base.metadata.create_all(engine)

            Session = sessionmaker(bind=engine)  # noqa: N806
            session = Session()

            bob = Person(name="Bob")
            session.add(bob)

            class fake_record_sql_queries:  # noqa: N801
                def __init__(self, *args, **kwargs):
                    self._ctx_mgr = record_sql_queries(*args, **kwargs)

                def __enter__(self):
                    self.span = self._ctx_mgr.__enter__()
                    self.span.start_timestamp = datetime(2024, 1, 1, microsecond=0)
                    return self.span

                def __exit__(self, type, value, traceback):
                    self._ctx_mgr.__exit__(type, value, traceback)
                    self.span.timestamp = datetime(2024, 1, 1, microsecond=101000)

            with mock.patch(
                "sentry_sdk.integrations.sqlalchemy.record_sql_queries",
                fake_record_sql_queries,
            ):
                assert session.query(Person).first() == bob

        (event,) = events

        for span in event["spans"]:
            if span.get("op") == "db" and span.get("description").startswith(
                "SELECT person"
            ):
                data = span.get("data", {})

                assert SPANDATA.CODE_LINENO in data
                assert SPANDATA.CODE_NAMESPACE in data
                assert SPANDATA.CODE_FILEPATH in data
                assert SPANDATA.CODE_FUNCTION in data

                assert type(data.get(SPANDATA.CODE_LINENO)) == int
                assert data.get(SPANDATA.CODE_LINENO) > 0
                assert (
                    data.get(SPANDATA.CODE_NAMESPACE)
                    == "tests.integrations.sqlalchemy.test_sqlalchemy"
                )
                assert data.get(SPANDATA.CODE_FILEPATH).endswith(
                    "tests/integrations/sqlalchemy/test_sqlalchemy.py"
                )

                is_relative_path = data.get(SPANDATA.CODE_FILEPATH)[0] != os.sep
                assert is_relative_path

                assert (
                    data.get(SPANDATA.CODE_FUNCTION)
                    == "test_query_source_if_duration_over_threshold"
                )
                break
        else:
            raise AssertionError("No db span found")


@pytest.mark.parametrize("span_streaming", [True, False])
def test_span_origin(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[SqlalchemyIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    engine = create_engine(
        "sqlite:///:memory:", connect_args={"check_same_thread": False}
    )
    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="custom parent"):
            with engine.connect() as con:
                con.execute(text("SELECT 0"))

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        assert spans[0]["attributes"]["sentry.origin"] == "auto.db.sqlalchemy"
        assert spans[1]["attributes"]["sentry.origin"] == "manual"
    else:
        events = capture_events()
        with start_transaction(name="foo"):
            with engine.connect() as con:
                con.execute(text("SELECT 0"))

        (event,) = events

        assert event["contexts"]["trace"]["origin"] == "manual"
        assert event["spans"][0]["origin"] == "auto.db.sqlalchemy"



import datetime
import os
import socket
import time
from http.client import HTTPConnection, HTTPSConnection
from http.server import BaseHTTPRequestHandler, HTTPServer
from socket import SocketIO
from threading import Thread
from unittest import mock
from urllib.error import HTTPError
from urllib.request import urlopen

import pytest

import sentry_sdk
from sentry_sdk import capture_message, continue_trace, start_transaction
from sentry_sdk.consts import MATCH_ALL, SPANDATA
from sentry_sdk.integrations.stdlib import StdlibIntegration
from tests.conftest import ApproxDict, create_mock_http_server, get_free_port

PORT = create_mock_http_server()


class MockProxyRequestHandler(BaseHTTPRequestHandler):
    def do_CONNECT(self):
        self.send_response(200, "Connection Established")
        self.end_headers()

        self.rfile.readline()

        response = b"HTTP/1.1 200 OK\r\nContent-Length: 0\r\n\r\n"
        self.wfile.write(response)
        self.wfile.flush()


def create_mock_proxy_server():
    proxy_port = get_free_port()
    proxy_server = HTTPServer(("localhost", proxy_port), MockProxyRequestHandler)
    proxy_thread = Thread(target=proxy_server.serve_forever)
    proxy_thread.daemon = True
    proxy_thread.start()
    return proxy_port


PROXY_PORT = create_mock_proxy_server()

CHUNK_DELAY = 0.1
NUM_CHUNKS = 3


class ChunkedResponseHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        self.send_response(200)
        self.send_header("Transfer-Encoding", "chunked")
        self.end_headers()
        for _ in range(NUM_CHUNKS):
            chunk = b"x" * 100
            self.wfile.write(f"{len(chunk):x}\r\n".encode() + chunk + b"\r\n")
            self.wfile.flush()
            time.sleep(CHUNK_DELAY)
        self.wfile.write(b"0\r\n\r\n")

    def log_message(self, *args):
        pass


def create_chunked_server():
    port = get_free_port()
    server = HTTPServer(("localhost", port), ChunkedResponseHandler)
    thread = Thread(target=server.serve_forever)
    thread.daemon = True
    thread.start()
    return port


CHUNKED_PORT = create_chunked_server()


def test_crumb_capture(sentry_init, capture_events):
    sentry_init(integrations=[StdlibIntegration()], send_default_pii=True)
    events = capture_events()

    url = "http://localhost:{}/some/random/url".format(PORT)
    urlopen(url)

    capture_message("Testing!")

    (event,) = events
    (crumb,) = event["breadcrumbs"]["values"]

    assert crumb["type"] == "http"
    assert crumb["category"] == "httplib"
    assert crumb["data"] == ApproxDict(
        {
            "url": url,
            SPANDATA.HTTP_METHOD: "GET",
            SPANDATA.HTTP_STATUS_CODE: 200,
            "reason": "OK",
            SPANDATA.HTTP_FRAGMENT: "",
            SPANDATA.HTTP_QUERY: "",
        }
    )


@pytest.mark.parametrize(
    "status_code,level",
    [
        (200, None),
        (301, None),
        (403, "warning"),
        (405, "warning"),
        (500, "error"),
    ],
)
def test_crumb_capture_client_error(sentry_init, capture_events, status_code, level):
    sentry_init(integrations=[StdlibIntegration()], send_default_pii=True)
    events = capture_events()

    url = f"http://localhost:{PORT}/status/{status_code}"  # noqa:E231
    try:
        urlopen(url)
    except HTTPError:
        pass

    capture_message("Testing!")

    (event,) = events
    (crumb,) = event["breadcrumbs"]["values"]

    assert crumb["type"] == "http"
    assert crumb["category"] == "httplib"

    if level is None:
        assert "level" not in crumb
    else:
        assert crumb["level"] == level

    assert crumb["data"] == ApproxDict(
        {
            "url": url,
            SPANDATA.HTTP_METHOD: "GET",
            SPANDATA.HTTP_STATUS_CODE: status_code,
            SPANDATA.HTTP_FRAGMENT: "",
            SPANDATA.HTTP_QUERY: "",
        }
    )


def test_crumb_capture_hint(sentry_init, capture_events):
    def before_breadcrumb(crumb, hint):
        crumb["data"]["extra"] = "foo"
        return crumb

    sentry_init(
        integrations=[StdlibIntegration()],
        before_breadcrumb=before_breadcrumb,
        send_default_pii=True,
    )
    events = capture_events()

    url = "http://localhost:{}/some/random/url".format(PORT)
    urlopen(url)

    capture_message("Testing!")

    (event,) = events
    (crumb,) = event["breadcrumbs"]["values"]
    assert crumb["type"] == "http"
    assert crumb["category"] == "httplib"
    assert crumb["data"] == ApproxDict(
        {
            "url": url,
            SPANDATA.HTTP_METHOD: "GET",
            SPANDATA.HTTP_STATUS_CODE: 200,
            "reason": "OK",
            "extra": "foo",
            SPANDATA.HTTP_FRAGMENT: "",
            SPANDATA.HTTP_QUERY: "",
        }
    )


@pytest.mark.parametrize("span_streaming", [True, False])
def test_empty_realurl(
    sentry_init,
    span_streaming,
):
    """
    Ensure that after using sentry_sdk.init you can putrequest a
    None url.
    """

    sentry_init(
        dsn="",
        trace_lifecycle="stream" if span_streaming else "static",
    )
    HTTPConnection("localhost", port=PORT).putrequest("POST", None)


def test_httplib_misuse(sentry_init, capture_events, request):
    """HTTPConnection.getresponse must be called after every call to
    HTTPConnection.request. However, if somebody does not abide by
    this contract, we still should handle this gracefully and not
    send mixed breadcrumbs.

    Test whether our breadcrumbs are coherent when somebody uses HTTPConnection
    wrongly.
    """

    sentry_init(send_default_pii=True)
    events = capture_events()

    conn = HTTPConnection("localhost", PORT)

    # make sure we release the resource, even if the test fails
    request.addfinalizer(conn.close)

    conn.request("GET", "/200")

    with pytest.raises(Exception):  # noqa: B017
        # This raises an exception, because we didn't call `getresponse` for
        # the previous request yet.
        #
        # This call should not affect our breadcrumb.
        conn.request("POST", "/200")

    response = conn.getresponse()
    assert response._method == "GET"

    capture_message("Testing!")

    (event,) = events
    (crumb,) = event["breadcrumbs"]["values"]

    assert crumb["type"] == "http"
    assert crumb["category"] == "httplib"
    assert crumb["data"] == ApproxDict(
        {
            "url": "http://localhost:{}/200".format(PORT),
            SPANDATA.HTTP_METHOD: "GET",
            SPANDATA.HTTP_STATUS_CODE: 200,
            "reason": "OK",
            SPANDATA.HTTP_FRAGMENT: "",
            SPANDATA.HTTP_QUERY: "",
        }
    )


@pytest.mark.parametrize("span_streaming", [True, False])
def test_outgoing_trace_headers(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    already_patched_getresponse = HTTPSConnection.getresponse
    request_headers = {}

    class HTTPSConnectionRecordingRequestHeaders(HTTPSConnection):
        def send(self, *args, **kwargs) -> None:
            request_str = args[0]
            for line in request_str.decode("utf-8").split("\r\n")[1:]:
                if line:
                    key, val = line.split(": ")
                    request_headers[key] = val

            server_sock, client_sock = socket.socketpair()
            server_sock.sendall(b"HTTP/1.1 200 OK\r\n\r\n")
            server_sock.close()
            self.sock = client_sock

        def getresponse(self, *args, **kwargs):
            return already_patched_getresponse(self, *args, **kwargs)

    events = capture_events()

    headers = {
        "sentry-trace": "771a43a4192642f0b136d5159a501700-1234567890abcdef-1",
        "baggage": (
            "other-vendor-value-1=foo;bar;baz,"
            "sentry-trace_id=771a43a4192642f0b136d5159a501700,"
            "sentry-public_key=49d0f7386ad645858ae85020e393bef3,"
            "sentry-sample_rate=0.01337,"
            "sentry-user_id=Am%C3%A9lie,"
            "sentry-sample_rand=0.000005,"
            "sentry-sampled=true,"
            "other-vendor-value-2=foo;bar;"
        ),
    }

    if span_streaming:
        items = capture_items("span")
        sentry_sdk.traces.continue_trace(headers)

        with sentry_sdk.traces.start_span(
            name="/interactions/other-dogs/new-dog",
            attributes={
                "sentry.op": "greeting.sniff",
            },
        ):
            connection = HTTPSConnectionRecordingRequestHeaders("localhost", port=PORT)
            connection.request("GET", "/top-chasers")
            connection.getresponse()

        sentry_sdk.flush()
        request_span = next(item.payload for item in items if item.type == "span")
        expected_sentry_trace = "{trace_id}-{parent_span_id}-{sampled}".format(
            trace_id=request_span["trace_id"],
            parent_span_id=request_span["span_id"],
            sampled=1,
        )

        expected_outgoing_baggage = (
            "sentry-trace_id=771a43a4192642f0b136d5159a501700,"
            "sentry-public_key=49d0f7386ad645858ae85020e393bef3,"
            "sentry-sample_rate=0.01337,"
            "sentry-user_id=Am%C3%A9lie,"
            "sentry-sample_rand=0.000005,"
            "sentry-sampled=true"
        )
    else:
        events = capture_events()
        transaction = continue_trace(headers)

        with start_transaction(
            transaction=transaction,
            name="/interactions/other-dogs/new-dog",
            op="greeting.sniff",
            trace_id="12312012123120121231201212312012",
        ) as transaction:
            connection = HTTPSConnectionRecordingRequestHeaders("localhost", port=PORT)
            connection.request("GET", "/top-chasers")
            connection.getresponse()

        (event,) = events
        request_span = event["spans"][-1]
        expected_sentry_trace = "{trace_id}-{parent_span_id}-{sampled}".format(
            trace_id=event["contexts"]["trace"]["trace_id"],
            parent_span_id=request_span["span_id"],
            sampled=1,
        )

        # Note: the sample rate here is actually wrong. It's fixed in the
        # streaming path
        expected_outgoing_baggage = (
            "sentry-trace_id=771a43a4192642f0b136d5159a501700,"
            "sentry-public_key=49d0f7386ad645858ae85020e393bef3,"
            "sentry-sample_rate=1.0,"
            "sentry-user_id=Am%C3%A9lie,"
            "sentry-sample_rand=0.000005,"
            "sentry-sampled=true"
        )

    assert request_headers["sentry-trace"] == expected_sentry_trace
    assert request_headers["baggage"] == expected_outgoing_baggage


@pytest.mark.parametrize("span_streaming", [True, False])
def test_outgoing_trace_headers_head_sdk(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        traces_sample_rate=0.5,
        release="foo",
        trace_lifecycle="stream" if span_streaming else "static",
    )

    already_patched_getresponse = HTTPSConnection.getresponse
    request_headers = {}

    class HTTPSConnectionRecordingRequestHeaders(HTTPSConnection):
        def send(self, *args, **kwargs) -> None:
            request_str = args[0]
            for line in request_str.decode("utf-8").split("\r\n")[1:]:
                if line:
                    key, val = line.split(": ")
                    request_headers[key] = val

            server_sock, client_sock = socket.socketpair()
            server_sock.sendall(b"HTTP/1.1 200 OK\r\n\r\n")
            server_sock.close()
            self.sock = client_sock

        def getresponse(self, *args, **kwargs):
            return already_patched_getresponse(self, *args, **kwargs)

    if span_streaming:
        items = capture_items("span")

        sentry_sdk.traces.continue_trace({})

        with mock.patch(
            "sentry_sdk.tracing_utils.Random.randrange", return_value=250000
        ):
            with sentry_sdk.traces.start_span(name="Head SDK tx"):
                connection = HTTPSConnectionRecordingRequestHeaders(
                    "localhost", port=PORT
                )
                connection.request("GET", "/top-chasers")
                connection.getresponse()

        sentry_sdk.flush()
        request_span = next(item.payload for item in items if item.type == "span")
        expected_sentry_trace = "{trace_id}-{parent_span_id}-{sampled}".format(
            trace_id=request_span["trace_id"],
            parent_span_id=request_span["span_id"],
            sampled=1,
        )

        expected_outgoing_baggage = (
            "sentry-trace_id=%s,"
            "sentry-sample_rand=0.250000,"
            "sentry-environment=production,"
            "sentry-release=foo,"
            "sentry-transaction=Head%%20SDK%%20tx,"
            "sentry-sample_rate=0.5,"
            "sentry-sampled=true"
        ) % request_span["trace_id"]
    else:
        events = capture_events()

        with mock.patch(
            "sentry_sdk.tracing_utils.Random.randrange", return_value=250000
        ):
            transaction = continue_trace({})

        with start_transaction(
            transaction=transaction, name="Head SDK tx"
        ) as transaction:
            connection = HTTPSConnectionRecordingRequestHeaders("localhost", port=PORT)
            connection.request("GET", "/top-chasers")
            connection.getresponse()

        (event,) = events
        request_span = event["spans"][-1]
        expected_sentry_trace = "{trace_id}-{parent_span_id}-{sampled}".format(
            trace_id=event["contexts"]["trace"]["trace_id"],
            parent_span_id=request_span["span_id"],
            sampled=1,
        )

        expected_outgoing_baggage = (
            "sentry-trace_id=%s,"
            "sentry-sample_rand=0.250000,"
            "sentry-environment=production,"
            "sentry-release=foo,"
            "sentry-sample_rate=0.5,"
            "sentry-sampled=%s"
        ) % (transaction.trace_id, "true" if transaction.sampled else "false")

    assert request_headers["sentry-trace"] == expected_sentry_trace
    assert request_headers["baggage"] == expected_outgoing_baggage


def test_outgoing_trace_headers_span_streaming_no_current_span(sentry_init):
    """
    With span streaming enabled and no active span, trace propagation headers
    should still be attached to outgoing requests, propagated from the scope's
    propagation context.
    """
    sentry_init(
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    already_patched_getresponse = HTTPSConnection.getresponse
    request_headers = {}

    class HTTPSConnectionRecordingRequestHeaders(HTTPSConnection):
        def send(self, *args, **kwargs) -> None:
            request_str = args[0]
            for line in request_str.decode("utf-8").split("\r\n")[1:]:
                if line:
                    key, val = line.split(": ")
                    request_headers[key] = val

            server_sock, client_sock = socket.socketpair()
            server_sock.sendall(b"HTTP/1.1 200 OK\r\n\r\n")
            server_sock.close()
            self.sock = client_sock

        def getresponse(self, *args, **kwargs):
            return already_patched_getresponse(self, *args, **kwargs)

    headers = {
        "sentry-trace": "771a43a4192642f0b136d5159a501700-1234567890abcdef-1",
        "baggage": (
            "sentry-trace_id=771a43a4192642f0b136d5159a501700,"
            "sentry-public_key=49d0f7386ad645858ae85020e393bef3,"
            "sentry-sample_rate=0.01337,"
            "sentry-sample_rand=0.000005,"
            "sentry-sampled=true"
        ),
    }

    # Seed the scope's propagation context, but do NOT start a span.
    sentry_sdk.traces.continue_trace(headers)
    assert sentry_sdk.traces.get_current_span() is None

    connection = HTTPSConnectionRecordingRequestHeaders("localhost", port=PORT)
    connection.request("GET", "/top-chasers")
    connection.getresponse()

    # Trace is still propagated, carrying the trace_id from the propagation
    # context. The span id segment is generated for the propagation context
    # (there is no active span), so we assert the stable trace_id prefix.
    assert request_headers["sentry-trace"].startswith(
        "771a43a4192642f0b136d5159a501700-"
    )

    # The outgoing baggage is fully deterministic: it is the frozen incoming
    # baggage from the propagation context.
    expected_outgoing_baggage = (
        "sentry-trace_id=771a43a4192642f0b136d5159a501700,"
        "sentry-public_key=49d0f7386ad645858ae85020e393bef3,"
        "sentry-sample_rate=0.01337,"
        "sentry-sample_rand=0.000005,"
        "sentry-sampled=true"
    )
    assert request_headers["baggage"] == expected_outgoing_baggage


@pytest.mark.parametrize(
    "trace_propagation_targets,host,path,trace_propagated",
    [
        [
            [],
            "example.com",
            "/",
            False,
        ],
        [
            None,
            "example.com",
            "/",
            False,
        ],
        [
            [MATCH_ALL],
            "example.com",
            "/",
            True,
        ],
        [
            ["https://example.com/"],
            "example.com",
            "/",
            True,
        ],
        [
            ["https://example.com/"],
            "example.com",
            "",
            False,
        ],
        [
            ["https://example.com"],
            "example.com",
            "",
            True,
        ],
        [
            ["https://example.com", r"https?:\/\/[\w\-]+(\.[\w\-]+)+\.net"],
            "example.net",
            "",
            False,
        ],
        [
            ["https://example.com", r"https?:\/\/[\w\-]+(\.[\w\-]+)+\.net"],
            "good.example.net",
            "",
            True,
        ],
        [
            ["https://example.com", r"https?:\/\/[\w\-]+(\.[\w\-]+)+\.net"],
            "good.example.net",
            "/some/thing",
            True,
        ],
    ],
)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_option_trace_propagation_targets(
    sentry_init,
    trace_propagation_targets,
    host,
    path,
    trace_propagated,
    span_streaming,
):
    sentry_init(
        trace_propagation_targets=trace_propagation_targets,
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    already_patched_getresponse = HTTPSConnection.getresponse

    request_headers = {}

    class HTTPSConnectionRecordingRequestHeaders(HTTPSConnection):
        def send(self, *args, **kwargs) -> None:
            request_str = args[0]
            for line in request_str.decode("utf-8").split("\r\n")[1:]:
                if line:
                    key, val = line.split(": ")
                    request_headers[key] = val

            server_sock, client_sock = socket.socketpair()
            server_sock.sendall(b"HTTP/1.1 200 OK\r\n\r\n")
            server_sock.close()
            self.sock = client_sock

        def getresponse(self, *args, **kwargs):
            return already_patched_getresponse(self, *args, **kwargs)

    headers = {
        "baggage": (
            "sentry-trace_id=771a43a4192642f0b136d5159a501700, "
            "sentry-public_key=49d0f7386ad645858ae85020e393bef3, sentry-sample_rate=0.01337, "
        )
    }

    if span_streaming:
        sentry_sdk.traces.continue_trace(headers)

        with sentry_sdk.traces.start_span(
            name="/interactions/other-dogs/new-dog",
            attributes={
                "sentry.op": "greeting.sniff",
            },
        ):
            connection = HTTPSConnectionRecordingRequestHeaders(host)
            connection.request("GET", path)
            connection.getresponse()
    else:
        transaction = continue_trace(headers)

        with start_transaction(
            transaction=transaction,
            name="/interactions/other-dogs/new-dog",
            op="greeting.sniff",
            trace_id="12312012123120121231201212312012",
        ) as transaction:
            connection = HTTPSConnectionRecordingRequestHeaders(host)
            connection.request("GET", path)
            connection.getresponse()

    if trace_propagated:
        assert "sentry-trace" in request_headers
        assert "baggage" in request_headers
    else:
        assert "sentry-trace" not in request_headers
        assert "baggage" not in request_headers


@pytest.mark.parametrize("span_streaming", [True, False])
def test_request_source_disabled(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_options = {
        "traces_sample_rate": 1.0,
        "enable_http_request_source": False,
        "http_request_source_threshold_ms": 0,
    }

    sentry_init(
        **sentry_options,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            conn = HTTPConnection("localhost", port=PORT)
            conn.request("GET", "/foo")
            conn.getresponse()

        sentry_sdk.flush()
        span = next(item.payload for item in items if item.type == "span")
        assert span["name"].startswith("GET")

        attributes = span["attributes"]

        assert SPANDATA.CODE_LINE_NUMBER not in attributes
        assert SPANDATA.CODE_NAMESPACE not in attributes
        assert SPANDATA.CODE_FILE_PATH not in attributes
        assert SPANDATA.CODE_FUNCTION not in attributes
    else:
        events = capture_events()

        with start_transaction(name="foo"):
            conn = HTTPConnection("localhost", port=PORT)
            conn.request("GET", "/foo")
            conn.getresponse()

        (event,) = events

        span = event["spans"][-1]
        assert span["description"].startswith("GET")

        data = span.get("data", {})

        assert SPANDATA.CODE_LINENO not in data
        assert SPANDATA.CODE_NAMESPACE not in data
        assert SPANDATA.CODE_FILEPATH not in data
        assert SPANDATA.CODE_FUNCTION not in data


@pytest.mark.parametrize("enable_http_request_source", [None, True])
@pytest.mark.parametrize("span_streaming", [True, False])
def test_request_source_enabled(
    sentry_init,
    capture_events,
    capture_items,
    enable_http_request_source,
    span_streaming,
):
    sentry_options = {
        "traces_sample_rate": 1.0,
        "http_request_source_threshold_ms": 0,
        "trace_lifecycle": "stream" if span_streaming else "static",
    }

    if enable_http_request_source is not None:
        sentry_options["enable_http_request_source"] = enable_http_request_source

    if span_streaming:
        sentry_init(
            **sentry_options,
        )

        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            conn = HTTPConnection("localhost", port=PORT)
            conn.request("GET", "/foo")
            conn.getresponse()

        sentry_sdk.flush()
        span = next(item.payload for item in items if item.type == "span")
        assert span["name"].startswith("GET")

        attributes = span["attributes"]

        assert SPANDATA.CODE_LINE_NUMBER in attributes
        assert SPANDATA.CODE_NAMESPACE in attributes
        assert SPANDATA.CODE_FILE_PATH in attributes
        assert SPANDATA.CODE_FUNCTION in attributes
    else:
        sentry_init(**sentry_options)

        events = capture_events()

        with start_transaction(name="foo"):
            conn = HTTPConnection("localhost", port=PORT)
            conn.request("GET", "/foo")
            conn.getresponse()

        (event,) = events

        span = event["spans"][-1]
        assert span["description"].startswith("GET")

        data = span.get("data", {})

        assert SPANDATA.CODE_LINENO in data
        assert SPANDATA.CODE_NAMESPACE in data
        assert SPANDATA.CODE_FILEPATH in data
        assert SPANDATA.CODE_FUNCTION in data


@pytest.mark.parametrize("span_streaming", [True, False])
def test_request_source(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        traces_sample_rate=1.0,
        enable_http_request_source=True,
        http_request_source_threshold_ms=0,
        trace_lifecycle="stream" if span_streaming else "static",
    )
    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            conn = HTTPConnection("localhost", port=PORT)
            conn.request("GET", "/foo")
            conn.getresponse()

        sentry_sdk.flush()
        span = next(item.payload for item in items if item.type == "span")
        assert span["name"].startswith("GET")

        attributes = span["attributes"]

        assert SPANDATA.CODE_LINE_NUMBER in attributes
        assert SPANDATA.CODE_NAMESPACE in attributes
        assert SPANDATA.CODE_FILE_PATH in attributes
        assert SPANDATA.CODE_FUNCTION in attributes

        assert type(attributes.get(SPANDATA.CODE_LINE_NUMBER)) == int
        assert attributes.get(SPANDATA.CODE_LINE_NUMBER) > 0
        assert (
            attributes.get(SPANDATA.CODE_NAMESPACE)
            == "tests.integrations.stdlib.test_httplib"
        )
        assert attributes.get(SPANDATA.CODE_FILE_PATH).endswith(
            "tests/integrations/stdlib/test_httplib.py"
        )

        is_relative_path = attributes.get(SPANDATA.CODE_FILE_PATH)[0] != os.sep
        assert is_relative_path

        assert attributes.get(SPANDATA.CODE_FUNCTION) == "test_request_source"
    else:
        events = capture_events()

        with start_transaction(name="foo"):
            conn = HTTPConnection("localhost", port=PORT)
            conn.request("GET", "/foo")
            conn.getresponse()

        (event,) = events

        span = event["spans"][-1]
        assert span["description"].startswith("GET")

        data = span.get("data", {})

        assert SPANDATA.CODE_LINENO in data
        assert SPANDATA.CODE_NAMESPACE in data
        assert SPANDATA.CODE_FILEPATH in data
        assert SPANDATA.CODE_FUNCTION in data

        assert type(data.get(SPANDATA.CODE_LINENO)) == int
        assert data.get(SPANDATA.CODE_LINENO) > 0
        assert (
            data.get(SPANDATA.CODE_NAMESPACE)
            == "tests.integrations.stdlib.test_httplib"
        )
        assert data.get(SPANDATA.CODE_FILEPATH).endswith(
            "tests/integrations/stdlib/test_httplib.py"
        )

        is_relative_path = data.get(SPANDATA.CODE_FILEPATH)[0] != os.sep
        assert is_relative_path

        assert data.get(SPANDATA.CODE_FUNCTION) == "test_request_source"


@pytest.mark.parametrize("span_streaming", [True, False])
def test_request_source_with_module_in_search_path(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    """
    Test that request source is relative to the path of the module it ran in
    """
    sentry_init(
        traces_sample_rate=1.0,
        enable_http_request_source=True,
        http_request_source_threshold_ms=0,
        trace_lifecycle="stream" if span_streaming else "static",
    )
    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            from httplib_helpers.helpers import get_request_with_connection

            conn = HTTPConnection("localhost", port=PORT)
            get_request_with_connection(conn, "/foo")

        sentry_sdk.flush()
        span = next(item.payload for item in items if item.type == "span")
        assert span["name"].startswith("GET")

        attributes = span["attributes"]

        assert SPANDATA.CODE_LINE_NUMBER in attributes
        assert SPANDATA.CODE_NAMESPACE in attributes
        assert SPANDATA.CODE_FILE_PATH in attributes
        assert SPANDATA.CODE_FUNCTION in attributes

        assert type(attributes.get(SPANDATA.CODE_LINE_NUMBER)) == int
        assert attributes.get(SPANDATA.CODE_LINE_NUMBER) > 0
        assert attributes.get(SPANDATA.CODE_NAMESPACE) == "httplib_helpers.helpers"
        assert attributes.get(SPANDATA.CODE_FILE_PATH) == "httplib_helpers/helpers.py"

        is_relative_path = attributes.get(SPANDATA.CODE_FILE_PATH)[0] != os.sep
        assert is_relative_path

        assert attributes.get(SPANDATA.CODE_FUNCTION) == "get_request_with_connection"
    else:
        events = capture_events()

        with start_transaction(name="foo"):
            from httplib_helpers.helpers import get_request_with_connection

            conn = HTTPConnection("localhost", port=PORT)
            get_request_with_connection(conn, "/foo")

        (event,) = events

        span = event["spans"][-1]
        assert span["description"].startswith("GET")

        data = span.get("data", {})

        assert SPANDATA.CODE_LINENO in data
        assert SPANDATA.CODE_NAMESPACE in data
        assert SPANDATA.CODE_FILEPATH in data
        assert SPANDATA.CODE_FUNCTION in data

        assert type(data.get(SPANDATA.CODE_LINENO)) == int
        assert data.get(SPANDATA.CODE_LINENO) > 0
        assert data.get(SPANDATA.CODE_NAMESPACE) == "httplib_helpers.helpers"
        assert data.get(SPANDATA.CODE_FILEPATH) == "httplib_helpers/helpers.py"

        is_relative_path = data.get(SPANDATA.CODE_FILEPATH)[0] != os.sep
        assert is_relative_path

        assert data.get(SPANDATA.CODE_FUNCTION) == "get_request_with_connection"


@pytest.mark.parametrize("span_streaming", [True, False])
def test_no_request_source_if_duration_too_short(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        traces_sample_rate=1.0,
        enable_http_request_source=True,
        http_request_source_threshold_ms=100,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    add_http_request_source = sentry_sdk.tracing_utils.add_http_request_source

    def add_http_request_source_with_pinned_timestamps(span):
        if span_streaming:
            span._start_timestamp = datetime.datetime(2024, 1, 1, microsecond=0)
            span._end_timestamp = datetime.datetime(2024, 1, 1, microsecond=99999)
            result = add_http_request_source(span)
            span._end_timestamp = None
            return result
        else:
            span.start_timestamp = datetime.datetime(2024, 1, 1, microsecond=0)
            span.timestamp = datetime.datetime(2024, 1, 1, microsecond=99999)
            return add_http_request_source(span)

    if span_streaming:
        items = capture_items("span")
        with mock.patch(
            "sentry_sdk.integrations.stdlib.add_http_request_source",
            add_http_request_source_with_pinned_timestamps,
        ):
            with sentry_sdk.traces.start_span(name="foo"):
                conn = HTTPConnection("localhost", port=PORT)
                conn.request("GET", "/foo")
                conn.getresponse()

        sentry_sdk.flush()
        span = next(item.payload for item in items if item.type == "span")
        assert span["name"].startswith("GET")

        attributes = span["attributes"]

        assert SPANDATA.CODE_LINE_NUMBER not in attributes
        assert SPANDATA.CODE_NAMESPACE not in attributes
        assert SPANDATA.CODE_FILE_PATH not in attributes
        assert SPANDATA.CODE_FUNCTION not in attributes
    else:
        events = capture_events()

        with mock.patch(
            "sentry_sdk.integrations.stdlib.add_http_request_source",
            add_http_request_source_with_pinned_timestamps,
        ):
            with start_transaction(name="foo"):
                conn = HTTPConnection("localhost", port=PORT)
                conn.request("GET", "/foo")
                conn.getresponse()

        (event,) = events

        span = event["spans"][-1]
        assert span["description"].startswith("GET")

        data = span.get("data", {})

        assert SPANDATA.CODE_LINENO not in data
        assert SPANDATA.CODE_NAMESPACE not in data
        assert SPANDATA.CODE_FILEPATH not in data
        assert SPANDATA.CODE_FUNCTION not in data


@pytest.mark.parametrize("span_streaming", [True, False])
def test_request_source_if_duration_over_threshold(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        traces_sample_rate=1.0,
        enable_http_request_source=True,
        http_request_source_threshold_ms=100,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    add_http_request_source = sentry_sdk.tracing_utils.add_http_request_source

    def add_http_request_source_with_pinned_timestamps(span):
        if span_streaming:
            span._start_timestamp = datetime.datetime(2024, 1, 1, microsecond=0)
            span._end_timestamp = datetime.datetime(2024, 1, 1, microsecond=100001)
            result = add_http_request_source(span)
            span._end_timestamp = None
            return result
        else:
            span.start_timestamp = datetime.datetime(2024, 1, 1, microsecond=0)
            span.timestamp = datetime.datetime(2024, 1, 1, microsecond=100001)
            return add_http_request_source(span)

    if span_streaming:
        items = capture_items("span")

        with mock.patch(
            "sentry_sdk.integrations.stdlib.add_http_request_source",
            add_http_request_source_with_pinned_timestamps,
        ):
            with sentry_sdk.traces.start_span(name="foo"):
                conn = HTTPConnection("localhost", port=PORT)
                conn.request("GET", "/foo")
                conn.getresponse()

        sentry_sdk.flush()
        span = next(item.payload for item in items if item.type == "span")
        assert span["name"].startswith("GET")

        attributes = span["attributes"]

        assert SPANDATA.CODE_LINE_NUMBER in attributes
        assert SPANDATA.CODE_NAMESPACE in attributes
        assert SPANDATA.CODE_FILE_PATH in attributes
        assert SPANDATA.CODE_FUNCTION in attributes

        assert type(attributes.get(SPANDATA.CODE_LINE_NUMBER)) == int
        assert attributes.get(SPANDATA.CODE_LINE_NUMBER) > 0
        assert (
            attributes.get(SPANDATA.CODE_NAMESPACE)
            == "tests.integrations.stdlib.test_httplib"
        )
        assert attributes.get(SPANDATA.CODE_FILE_PATH).endswith(
            "tests/integrations/stdlib/test_httplib.py"
        )

        is_relative_path = attributes.get(SPANDATA.CODE_FILE_PATH)[0] != os.sep
        assert is_relative_path

        assert (
            attributes.get(SPANDATA.CODE_FUNCTION)
            == "add_http_request_source_with_pinned_timestamps"
        )
    else:
        events = capture_events()

        with mock.patch(
            "sentry_sdk.integrations.stdlib.add_http_request_source",
            add_http_request_source_with_pinned_timestamps,
        ):
            with start_transaction(name="foo"):
                conn = HTTPConnection("localhost", port=PORT)
                conn.request("GET", "/foo")
                conn.getresponse()

        (event,) = events

        span = event["spans"][-1]
        assert span["description"].startswith("GET")

        data = span.get("data", {})

        assert SPANDATA.CODE_LINENO in data
        assert SPANDATA.CODE_NAMESPACE in data
        assert SPANDATA.CODE_FILEPATH in data
        assert SPANDATA.CODE_FUNCTION in data

        assert type(data.get(SPANDATA.CODE_LINENO)) == int
        assert data.get(SPANDATA.CODE_LINENO) > 0
        assert (
            data.get(SPANDATA.CODE_NAMESPACE)
            == "tests.integrations.stdlib.test_httplib"
        )
        assert data.get(SPANDATA.CODE_FILEPATH).endswith(
            "tests/integrations/stdlib/test_httplib.py"
        )

        is_relative_path = data.get(SPANDATA.CODE_FILEPATH)[0] != os.sep
        assert is_relative_path

        assert (
            data.get(SPANDATA.CODE_FUNCTION)
            == "add_http_request_source_with_pinned_timestamps"
        )


@pytest.mark.parametrize("span_streaming", [True, False])
def test_span_origin(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        traces_sample_rate=1.0,
        debug=True,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            conn = HTTPConnection("localhost", port=PORT)
            conn.request("GET", "/foo")
            conn.getresponse()

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        assert spans[1]["attributes"]["sentry.origin"] == "manual"

        assert spans[0]["attributes"]["sentry.op"] == "http.client"
        assert spans[0]["attributes"]["sentry.origin"] == "auto.http.stdlib.httplib"
    else:
        events = capture_events()

        with start_transaction(name="foo"):
            conn = HTTPConnection("localhost", port=PORT)
            conn.request("GET", "/foo")
            conn.getresponse()

        (event,) = events
        assert event["contexts"]["trace"]["origin"] == "manual"

        assert event["spans"][0]["op"] == "http.client"
        assert event["spans"][0]["origin"] == "auto.http.stdlib.httplib"


@pytest.mark.parametrize("span_streaming", [True, False])
def test_http_timeout(
    monkeypatch,
    sentry_init,
    capture_envelopes,
    capture_items,
    span_streaming,
):
    mock_readinto = mock.Mock(side_effect=TimeoutError)
    monkeypatch.setattr(SocketIO, "readinto", mock_readinto)

    sentry_init(
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if span_streaming:
        items = capture_items("span")

        with pytest.raises(TimeoutError):
            with sentry_sdk.traces.start_span(
                name="name",
                attributes={
                    "sentry.op": "op",
                },
            ):
                conn = HTTPConnection("localhost", port=PORT)
                conn.request("GET", "/bla")
                conn.getresponse()

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        assert len(spans) == 2
        span = spans[0]
        assert span["attributes"]["sentry.op"] == "http.client"
        assert span["name"] == f"GET http://localhost:{PORT}/bla"  # noqa: E231
    else:
        envelopes = capture_envelopes()

        with pytest.raises(TimeoutError):
            with start_transaction(op="op", name="name"):
                conn = HTTPConnection("localhost", port=PORT)
                conn.request("GET", "/bla")
                conn.getresponse()

        (transaction_envelope,) = envelopes
        transaction = transaction_envelope.get_transaction_event()
        assert len(transaction["spans"]) == 1

        span = transaction["spans"][0]
        assert span["op"] == "http.client"
        assert span["description"] == f"GET http://localhost:{PORT}/bla"  # noqa: E231


@pytest.mark.parametrize("tunnel_port", [8080, None])
@pytest.mark.parametrize("span_streaming", [True, False])
@pytest.mark.parametrize("send_default_pii", [True, False])
def test_proxy_http_tunnel(
    sentry_init,
    capture_events,
    capture_items,
    tunnel_port,
    span_streaming,
    send_default_pii,
):
    sentry_init(
        traces_sample_rate=1.0,
        send_default_pii=send_default_pii,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            conn = HTTPConnection("localhost", PROXY_PORT)
            conn.set_tunnel("api.example.com", tunnel_port)
            conn.request("GET", "/foo?bar=1")
            conn.getresponse()

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        (span,) = (
            span
            for span in spans
            if span["attributes"].get("sentry.origin") == "auto.http.stdlib.httplib"
        )

        port_modifier = f":{tunnel_port}" if tunnel_port else ""
        assert span["name"] == f"GET http://api.example.com{port_modifier}/foo"
        assert span["attributes"][SPANDATA.HTTP_REQUEST_METHOD] == "GET"
        assert span["attributes"][SPANDATA.NETWORK_PEER_ADDRESS] == "localhost"
        assert span["attributes"][SPANDATA.NETWORK_PEER_PORT] == PROXY_PORT

        if send_default_pii:
            assert (
                span["attributes"][SPANDATA.URL_FULL]
                == f"http://api.example.com{port_modifier}/foo"
            )
            assert span["attributes"][SPANDATA.URL_QUERY] == "bar=1"
        else:
            assert SPANDATA.URL_FULL not in span["attributes"]
            assert SPANDATA.URL_QUERY not in span["attributes"]
    else:
        events = capture_events()

        with start_transaction(name="test_transaction"):
            conn = HTTPConnection("localhost", PROXY_PORT)
            conn.set_tunnel("api.example.com", tunnel_port)
            conn.request("GET", "/foo?bar=1")
            conn.getresponse()

        (event,) = events
        (span,) = event["spans"]

        port_modifier = f":{tunnel_port}" if tunnel_port else ""
        assert span["description"] == f"GET http://api.example.com{port_modifier}/foo"
        assert span["data"][SPANDATA.HTTP_METHOD] == "GET"
        assert span["data"][SPANDATA.NETWORK_PEER_ADDRESS] == "localhost"
        assert span["data"][SPANDATA.NETWORK_PEER_PORT] == PROXY_PORT
        assert span["data"]["url"] == f"http://api.example.com{port_modifier}/foo"
        assert span["data"][SPANDATA.HTTP_QUERY] == "bar=1"


@pytest.mark.parametrize("span_streaming", [True, False])
def test_chunked_response_span_covers_body_read(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    min_expected_duration = CHUNK_DELAY * NUM_CHUNKS

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            conn = HTTPConnection("localhost", CHUNKED_PORT)
            conn.request("GET", "/chunked")
            response = conn.getresponse()
            response.read()

        sentry_sdk.flush()
        http_span, parent_span = [item.payload for item in items]

        duration = http_span["end_timestamp"] - http_span["start_timestamp"]
        assert duration >= min_expected_duration
    else:
        events = capture_events()

        with start_transaction(name="test_chunked"):
            conn = HTTPConnection("localhost", CHUNKED_PORT)
            conn.request("GET", "/chunked")
            response = conn.getresponse()
            response.read()

        (event,) = events
        (span,) = event["spans"]

        fmt = "%Y-%m-%dT%H:%M:%S.%fZ"
        start = datetime.datetime.strptime(span["start_timestamp"], fmt)
        end = datetime.datetime.strptime(span["timestamp"], fmt)
        duration = (end - start).total_seconds()
        assert duration >= min_expected_duration
